import datetime
import logging
from typing import Any

import pytest
import torch

from fme.core.coordinates import (
    HybridSigmaPressureCoordinate,
    LatLonCoordinates,
    NullVerticalCoordinate,
)
from fme.core.dataset.data_typing import VariableMetadata
from fme.core.dataset_info import (
    DatasetInfo,
    IncompatibleDatasetInfo,
    MissingDatasetInfo,
    get_keys_with_conflicts,
)
from fme.core.gridded_ops import LatLonOperations
from fme.core.mask_provider import MaskProvider
from fme.core.metrics import spherical_area_weights


@pytest.mark.parametrize(
    "dataset_info",
    [
        pytest.param(
            DatasetInfo(),
            id="empty",
        ),
        pytest.param(
            DatasetInfo(
                horizontal_coordinates=LatLonCoordinates(
                    lat=torch.arange(-4, 4), lon=torch.arange(16)
                ),
                vertical_coordinate=HybridSigmaPressureCoordinate(
                    ak=torch.arange(10),
                    bk=torch.arange(10),
                ),
                timestep=datetime.timedelta(hours=1),
            ),
            id="vertical_coordinate",
        ),
        pytest.param(
            DatasetInfo(
                horizontal_coordinates=LatLonCoordinates(
                    lat=torch.arange(-4, 4), lon=torch.arange(16)
                ),
                mask_provider=MaskProvider(masks={"mask_0": torch.ones(10, 10)}),
                timestep=datetime.timedelta(hours=1),
            ),
            id="mask_provider",
        ),
        pytest.param(
            DatasetInfo(
                variable_metadata={
                    "var_0": VariableMetadata(
                        "m",
                        "Variable 0",
                    ),
                },
            ),
            id="variable_metadata",
        ),
    ],
)
def test_dataset_info_round_trip(dataset_info: DatasetInfo):
    state = dataset_info.to_state()
    dataset_info_reloaded = DatasetInfo.from_state(state)
    assert dataset_info == dataset_info_reloaded


@pytest.mark.parametrize(
    "a, b",
    [
        pytest.param(
            DatasetInfo(),
            DatasetInfo(),
            id="empty",
        ),
        pytest.param(
            DatasetInfo(
                horizontal_coordinates=LatLonCoordinates(
                    lat=torch.arange(-4, 4),
                    lon=torch.arange(16),
                )
            ),
            DatasetInfo(
                horizontal_coordinates=LatLonCoordinates(
                    lat=torch.arange(-4, 4),
                    lon=torch.arange(16),
                )
            ),
            id="horizontal_coordinates_equal",
        ),
        pytest.param(
            DatasetInfo(),
            DatasetInfo(
                horizontal_coordinates=LatLonCoordinates(
                    lat=torch.arange(-4, 4),
                    lon=torch.arange(16),
                )
            ),
            id="horizontal_coordinates_missing_from_first",
        ),  # model without HorizontalCoordinates doesn't care what inference uses
        pytest.param(
            DatasetInfo(
                horizontal_coordinates=LatLonCoordinates(
                    lat=torch.arange(-4, 4),
                    lon=torch.arange(16),
                )
            ),
            DatasetInfo(),
            id="horizontal_coordinates_missing_from_second",
        ),
        pytest.param(
            DatasetInfo(
                vertical_coordinate=HybridSigmaPressureCoordinate(
                    ak=torch.arange(7), bk=torch.arange(7)
                )
            ),
            DatasetInfo(
                vertical_coordinate=HybridSigmaPressureCoordinate(
                    ak=torch.arange(7), bk=torch.arange(7)
                )
            ),
            id="vertical_coordinate_equal",
        ),
        pytest.param(
            DatasetInfo(
                vertical_coordinate=HybridSigmaPressureCoordinate(
                    ak=torch.arange(7), bk=torch.arange(7)
                )
            ),
            DatasetInfo(),
            id="vertical_coordinate_missing_from_second",
        ),
        pytest.param(
            DatasetInfo(),
            DatasetInfo(
                vertical_coordinate=HybridSigmaPressureCoordinate(
                    ak=torch.arange(7), bk=torch.arange(7)
                )
            ),
            id="vertical_coordinate_missing_from_first",
        ),
        pytest.param(
            DatasetInfo(
                vertical_coordinate=NullVerticalCoordinate(),
            ),
            DatasetInfo(
                vertical_coordinate=HybridSigmaPressureCoordinate(
                    ak=torch.arange(7), bk=torch.arange(7)
                )
            ),
            id="null_vertical_coordinate_first",
        ),
        pytest.param(
            DatasetInfo(
                mask_provider=MaskProvider(masks={"mask_0": torch.ones(10, 10)})
            ),
            DatasetInfo(
                mask_provider=MaskProvider(masks={"mask_0": torch.ones(10, 10)})
            ),
            id="mask_provider_equal",
        ),
        pytest.param(
            DatasetInfo(mask_provider=MaskProvider()),
            DatasetInfo(),
            id="mask_provider_missing_from_second",
        ),
        pytest.param(
            DatasetInfo(),
            DatasetInfo(mask_provider=MaskProvider()),
            id="mask_provider_missing_from_first",
        ),
        pytest.param(
            DatasetInfo(),
            DatasetInfo(timestep=datetime.timedelta(hours=1)),
            id="timestep_missing_from_first",
        ),  # a model trained with no timestep can work with any timestep
        pytest.param(
            DatasetInfo(),
            DatasetInfo(
                variable_metadata={
                    "var_0": VariableMetadata(
                        "m",
                        "Variable 0",
                    )
                }
            ),
            id="variable_metadata_missing_from_first",
        ),  # different variable metadata is allowed
    ],
)
def test_assert_compatible_with_compatible_dataset_info(a: DatasetInfo, b: DatasetInfo):
    a.assert_compatible_with(b)


@pytest.mark.parametrize(
    "a, b, msgs",
    [
        pytest.param(
            DatasetInfo(
                horizontal_coordinates=LatLonCoordinates(
                    lat=torch.arange(-4, 4),
                    lon=torch.arange(16),
                )
            ),
            DatasetInfo(
                horizontal_coordinates=LatLonCoordinates(
                    lat=torch.arange(-5, 3),
                    lon=torch.arange(16),
                )
            ),
            ["horizontal_coordinates"],
            id="horizontal_coordinates_values_differ",
        ),
        pytest.param(
            DatasetInfo(
                vertical_coordinate=HybridSigmaPressureCoordinate(
                    ak=torch.arange(7), bk=torch.arange(7)
                )
            ),
            DatasetInfo(
                vertical_coordinate=HybridSigmaPressureCoordinate(
                    ak=torch.arange(10), bk=torch.arange(10)
                )
            ),
            ["vertical_coordinate"],
            id="vertical_coordinate_values_differ",
        ),
        pytest.param(
            DatasetInfo(
                mask_provider=MaskProvider(masks={"mask_0": torch.ones(10, 10)})
            ),
            DatasetInfo(
                mask_provider=MaskProvider(masks={"mask_0": torch.zeros(10, 10)})
            ),
            ["mask_provider"],
            id="mask_provider_masks_differ",
        ),
        pytest.param(
            DatasetInfo(timestep=datetime.timedelta(hours=1)),
            DatasetInfo(timestep=datetime.timedelta(hours=2)),
            ["timestep"],
            id="timestep_values_differ",
        ),
        pytest.param(
            DatasetInfo(timestep=datetime.timedelta(hours=1)),
            DatasetInfo(),
            ["timestep"],
            id="timestep_missing_from_second",
        ),  # a model trained with a timestep cannot work with arbitrary timesteps
        pytest.param(
            DatasetInfo(
                horizontal_coordinates=LatLonCoordinates(
                    lat=torch.arange(-4, 4),
                    lon=torch.arange(16),
                ),
                timestep=datetime.timedelta(hours=1),
            ),
            DatasetInfo(
                horizontal_coordinates=LatLonCoordinates(
                    lat=torch.arange(-5, 3),
                    lon=torch.arange(16),
                ),
                timestep=datetime.timedelta(hours=2),
            ),
            ["horizontal_coordinates", "timestep"],
            id="multiple_values_differ",
        ),
    ],
)
def test_assert_compatible_with_incompatible_dataset_info(
    a: DatasetInfo, b: DatasetInfo, msgs: list[str]
):
    with pytest.raises(IncompatibleDatasetInfo) as exc_info:
        a.assert_compatible_with(b)
    for msg in msgs:
        assert msg in str(exc_info.value)
    non_messages = set(
        ["horizontal_coordinates", "vertical_coordinate", "timestep", "img_shape"]
    ).difference(msgs)
    for msg in non_messages:
        assert msg not in str(exc_info.value)


def test_get_keys_with_conflicts():
    dict_a = {
        "var_1": VariableMetadata(
            "m",
            "Variable 1",
        ),
    }
    dict_b = {
        "var_0": VariableMetadata(
            "m",
            "Variable 0",
        ),
        "var_1": VariableMetadata(
            "cm",
            "Variable 1",
        ),
    }
    keys_with_conflicts = get_keys_with_conflicts(dict_a, dict_b)
    assert len(keys_with_conflicts) == 1
    assert next(iter(keys_with_conflicts.keys())) == "var_1"
    assert keys_with_conflicts["var_1"] == (
        VariableMetadata("m", "Variable 1"),
        VariableMetadata("cm", "Variable 1"),
    )


def test_compatibility_logs_metadata_keys(caplog):
    dataset_info_a = DatasetInfo(
        variable_metadata={
            "var_1": VariableMetadata(
                "m",
                "Variable 1",
            ),
        }
    )
    dataset_info_b = DatasetInfo(
        variable_metadata={
            "var_1": VariableMetadata(
                "cm",
                "Variable 1's other name",
            ),
        }
    )
    with caplog.at_level(logging.WARNING):
        dataset_info_a.assert_compatible_with(dataset_info_b)
    assert (
        "DatasetInfo has different metadata from other DatasetInfo for key var_1"
        in caplog.text
    )


def test_backwards_compatibility_with_gridded_ops():
    """Previously DatasetInfo was initialized with img_shape and
    gridded_operations. Now we use horizontal_coordinates. This test ensures
    that old-style DatasetInfo can still be deserialized."""
    legacy_state: dict[str, Any] = {
        "img_shape": (4, 4),
        "gridded_operations": {
            "type": "LatLonOperations",
            "state": {"area_weights": torch.ones(4, 4)},
        },
    }
    dataset_info = DatasetInfo.from_state(legacy_state)
    with pytest.raises(MissingDatasetInfo, match="horizontal_coordinates"):
        dataset_info.horizontal_coordinates
    assert isinstance(dataset_info.gridded_operations, LatLonOperations)
    assert dataset_info.img_shape == (4, 4)


def test_dataset_info_raises_error_with_conflicting_inputs():
    coords = LatLonCoordinates(lat=torch.arange(-4, 4), lon=torch.arange(16))
    with pytest.raises(ValueError, match="provide both img_shape"):
        DatasetInfo(horizontal_coordinates=coords, img_shape=(10, 10))
    with pytest.raises(ValueError, match="provide both gridded_operations"):
        ops = LatLonOperations(area_weights=torch.ones(10, 10))
        DatasetInfo(horizontal_coordinates=coords, gridded_operations=ops)


def test_masked_gridded_ops():
    lon = torch.arange(4)
    lat = torch.arange(2)
    coords = LatLonCoordinates(lat=lat, lon=lon)
    mask_provider = MaskProvider(masks={"mask_0": torch.ones(10, 10)})
    dataset_info = DatasetInfo(
        horizontal_coordinates=coords,
        mask_provider=mask_provider,
    )
    assert dataset_info.horizontal_coordinates == coords
    assert dataset_info.mask_provider == mask_provider
    expected_gridded_ops = LatLonOperations(
        area_weights=spherical_area_weights(lat, len(lon)),
        mask_provider=mask_provider,
    )
    assert dataset_info.gridded_operations == expected_gridded_ops



import abc
import dataclasses
from collections.abc import Callable, Mapping

# we use Type to distinguish from type attr of PerturbationSelector
from typing import Any, ClassVar, Type  # noqa: UP035

import dacite
import numpy as np
import torch

from fme.core.registry.registry import Registry


@dataclasses.dataclass
class PerturbationConfig(abc.ABC):
    """
    Returns a perturbation function config class.
    """

    @classmethod
    def from_state(cls, state: Mapping[str, Any]) -> "PerturbationConfig":
        """
        Create a PerturbationSelector from a dictionary containing all the information
        needed to build a PerturbationConfig.
        """
        return dacite.from_dict(
            data_class=cls, data=state, config=dacite.Config(strict=True)
        )

    @abc.abstractmethod
    def apply_perturbation(
        self,
        data: torch.Tensor,
        lat: torch.Tensor,
        lon: torch.Tensor,
        ocean_fraction: torch.Tensor,
    ) -> None: ...


@dataclasses.dataclass
class PerturbationSelector:
    type: str
    config: Mapping[str, Any]
    registry: ClassVar[Registry[PerturbationConfig]] = Registry[PerturbationConfig]()

    def __post_init__(self):
        if not isinstance(self.registry, Registry):
            raise ValueError("PerturbationSelector.registry should not be set manually")

    @classmethod
    def register(
        cls, type_name
    ) -> Callable[[Type[PerturbationConfig]], Type[PerturbationConfig]]:  # noqa: UP006
        return cls.registry.register(type_name)

    def build(self) -> PerturbationConfig:
        return self.registry.get(self.type, self.config)

    @classmethod
    def get_available_types(cls):
        """This class method is used to expose all available types of Perturbations."""
        return cls(type="", config={}).registry._types.keys()


@dataclasses.dataclass
class SSTPerturbation:
    """
    Configuration for sea surface temperature perturbations
    applied to initial condition and forcing data.
    Currently, this is strictly applied to both.

    Parameters:
        sst: List of perturbation selectors for SST perturbations.
    """

    sst: list[PerturbationSelector]

    def __post_init__(self):
        self.perturbations: list[PerturbationConfig] = [
            perturbation.build() for perturbation in self.sst
        ]


def _get_ocean_mask(ocean_fraction: torch.Tensor, cutoff: float = 0.5) -> torch.Tensor:
    return ocean_fraction > cutoff


@PerturbationSelector.register("constant")
@dataclasses.dataclass
class ConstantConfig(PerturbationConfig):
    """
    Configuration for a constant perturbation.
    """

    amplitude: float = 1.0

    def apply_perturbation(
        self,
        data: torch.Tensor,
        lat: torch.Tensor,
        lon: torch.Tensor,
        ocean_fraction: torch.Tensor,
    ):
        ocean_mask = _get_ocean_mask(ocean_fraction)
        data[ocean_mask] += self.amplitude  # type: ignore


@PerturbationSelector.register("greens_function")
@dataclasses.dataclass
class GreensFunctionConfig(PerturbationConfig):
    """
    Configuration for a single sinusoidal patch of a Green's function perturbation.
    See equation 1 in Bloch‐Johnson, J., et al. (2024).

    Parameters:
        amplitude: The amplitude of the perturbation,
            maximum is reached at (lat_center, lon_center).
        lat_center: The latitude at the center of the patch in degrees.
        lon_center: The longitude at the center of the patch in degrees.
        lat_width: latitudinal width of the patch in degrees.
        lon_width: longitudinal width of the patch in degrees.
    """

    amplitude: float = 1.0
    lat_center: float = 0.0
    lon_center: float = 0.0
    lat_width: float = 10.0
    lon_width: float = 10.0

    def __post_init__(self):
        self._lat_center_rad = np.deg2rad(self.lat_center)
        self._lon_center_rad = np.deg2rad(self.lon_center)
        self._lat_width_rad = np.deg2rad(self.lat_width)
        self._lon_width_rad = np.deg2rad(self.lon_width)

    def _wrap_longitude_discontinuity(
        self,
        lon: torch.Tensor,
    ) -> tuple[torch.Tensor, torch.Tensor]:
        """
        Assume longitude is in the range [0, 360) degrees.
        If the patch crosses the discontinuity at 0/360 degrees,
        shift the longitude accordingly.
        """
        lon_min = self.lon_center - self.lon_width / 2.0
        lon_max = self.lon_center + self.lon_width / 2.0
        if lon_min < 0:
            lon_shifted = ((lon + 180) % 360) - 180
            lon_in_patch = (lon_shifted > lon_min) & (lon_shifted < lon_max)
        elif lon_max > 360:
            lon_in_patch = (lon > lon_min) | (lon < lon_max % 360)
            lon_shifted = ((lon + 180) % 360) + 180
        else:
            lon_in_patch = (lon > lon_min) & (lon < lon_max)
            lon_shifted = lon
        return lon_in_patch, lon_shifted

    def apply_perturbation(
        self,
        data: torch.Tensor,
        lat: torch.Tensor,
        lon: torch.Tensor,
        ocean_fraction: torch.Tensor,
    ):
        lat_in_patch = torch.abs(lat - self.lat_center) < self.lat_width / 2.0
        lon_in_patch, lon_shifted = self._wrap_longitude_discontinuity(lon)
        mask = lat_in_patch & lon_in_patch
        ocean_mask = _get_ocean_mask(ocean_fraction)
        perturbation = self.amplitude * (
            torch.cos(
                torch.pi
                / 2
                * (lat.deg2rad() - self._lat_center_rad)
                / (self._lat_width_rad / 2.0)
            )
            ** 2
            * torch.cos(
                torch.pi
                / 2
                * (lon_shifted.deg2rad() - self._lon_center_rad)
                / (self._lon_width_rad / 2.0)
            )
            ** 2
        )
        mask = mask.expand(data.shape)
        perturbation = perturbation.expand(data.shape)
        data[mask & ocean_mask] += perturbation[mask & ocean_mask]



import dataclasses

import torch

from fme.ace.aggregator.one_step.deterministic import (
    DeterministicTrainOutput,
    OneStepDeterministicAggregator,
)
from fme.ace.aggregator.one_step.ensemble import get_one_step_ensemble_aggregator
from fme.ace.stepper import TrainOutput
from fme.core.dataset_info import DatasetInfo
from fme.core.generics.aggregator import AggregatorABC
from fme.core.tensors import fold_ensemble_dim, fold_sized_ensemble_dim
from fme.core.typing_ import TensorMapping


class OneStepAggregator(AggregatorABC[TrainOutput]):
    """
    Aggregates statistics for the first timestep.

    To use, call `record_batch` on the results of each batch, then call
    `get_logs` to get a dictionary of statistics when you're done.
    """

    def __init__(
        self,
        dataset_info: DatasetInfo,
        save_diagnostics: bool = True,
        output_dir: str | None = None,
        loss_scaling: TensorMapping | None = None,
        log_snapshots: bool = True,
        log_mean_maps: bool = True,
    ):
        """
        Args:
            dataset_info: Dataset coordinates and metadata.
            save_diagnostics: Whether to save diagnostics.
            output_dir: Directory to write diagnostics to.
            loss_scaling: Dictionary of variables and their scaling factors
                used in loss computation.
            log_snapshots: Whether to include snapshots in diagnostics.
            log_mean_maps: Whether to include mean maps in diagnostics.
        """
        self._deterministic_aggregator = OneStepDeterministicAggregator(
            dataset_info=dataset_info,
            save_diagnostics=save_diagnostics,
            output_dir=output_dir,
            loss_scaling=loss_scaling,
            log_snapshots=log_snapshots,
            log_mean_maps=log_mean_maps,
        )
        self._ensemble_aggregator = get_one_step_ensemble_aggregator(
            gridded_operations=dataset_info.gridded_operations,
            log_mean_maps=log_mean_maps,
            target_time=1,
            metadata=dataset_info.variable_metadata,
        )
        self._ensemble_recorded = False

    @torch.no_grad()
    def record_batch(
        self,
        batch: TrainOutput,
    ):
        folded_gen_data, n_ensemble = fold_ensemble_dim(batch.gen_data)
        folded_target_data = fold_sized_ensemble_dim(batch.target_data, n_ensemble)
        self._deterministic_aggregator.record_batch(
            DeterministicTrainOutput(
                metrics=batch.metrics,
                gen_data=folded_gen_data,
                target_data=folded_target_data,
                normalize=batch.normalize,
            )
        )
        if n_ensemble > 1:
            self._ensemble_aggregator.record_batch(
                target_data=batch.target_data,
                gen_data=batch.gen_data,
                i_time_start=0,
            )
            self._ensemble_recorded = True

    @torch.no_grad()
    def get_logs(self, label: str):
        """
        Returns logs as can be reported to WandB.

        Args:
            label: Label to prepend to all log keys.
        """
        deterministic_logs = self._deterministic_aggregator.get_logs(label)
        if self._ensemble_recorded:
            stochastic_logs = self._ensemble_aggregator.get_logs(label)
            if len(set(deterministic_logs.keys()) & set(stochastic_logs.keys())) > 0:
                raise ValueError(
                    "Stochastic and deterministic logs have overlapping keys, "
                    f"stochastic logs: {stochastic_logs}, "
                    f"deterministic logs: {deterministic_logs}"
                )
            return {**deterministic_logs, **stochastic_logs}
        else:
            return deterministic_logs

    @torch.no_grad()
    def flush_diagnostics(self, subdir: str | None = None):
        self._deterministic_aggregator.flush_diagnostics(subdir)


@dataclasses.dataclass
class OneStepAggregatorConfig:
    """
    Configuration for the validation OneStepAggregator.

    Arguments:
        log_snapshots: Whether to log snapshot images.
        log_mean_maps: Whether to log mean map images.
    """

    log_snapshots: bool = True
    log_mean_maps: bool = True

    def build(
        self,
        dataset_info: DatasetInfo,
        save_diagnostics: bool = True,
        output_dir: str | None = None,
        loss_scaling: TensorMapping | None = None,
    ):
        return OneStepAggregator(
            dataset_info=dataset_info,
            save_diagnostics=save_diagnostics,
            output_dir=output_dir,
            loss_scaling=loss_scaling,
            log_snapshots=self.log_snapshots,
            log_mean_maps=self.log_mean_maps,
        )



import collections
import logging
from collections import namedtuple
from collections.abc import Mapping
from typing import Literal

import matplotlib.figure
import matplotlib.pyplot as plt
import numpy as np
import torch
import xarray as xr

from fme.core.metrics import quantile
from fme.core.typing_ import TensorMapping

EPSILON = 1.0e-6


def trim_zero_bins(
    counts: np.ndarray, bin_edges: np.ndarray
) -> tuple[np.ndarray, np.ndarray]:
    """
    Trim bins with zero counts from the edges of the histogram.

    Args:
        counts: 1D array of counts for each bin
        bin_edges: 1D array of bin edges

    Returns:
        counts, bin_edges: trimmed arrays
    """
    mask = counts > 0
    first_nonzero = np.argmax(mask)
    last_nonzero = len(mask) - np.argmax(mask[::-1])
    return (
        counts[first_nonzero:last_nonzero],
        bin_edges[first_nonzero : last_nonzero + 1],
    )


def histogram(
    arr: torch.Tensor, vmin: float, bin_size: float, n_bins: int
) -> torch.Tensor:
    """
    Compute histogram of a tensor.

    Args:
        arr: 1D tensor of values
        vmin: minimum value of the histogram
        bin_size: size of each bin
        n_bins: number of bins

    Returns:
        counts, bin_edges: arrays of counts and bin edges
    """
    # use np.bincount for speed
    counts = torch.bincount(((arr - vmin) / bin_size).int(), minlength=n_bins)
    return_counts = counts[:n_bins]
    return_counts[-1] += torch.sum(counts[n_bins:])
    return return_counts


class DynamicHistogram:
    """
    A histogram that dynamically bins values into a fixed number of bins
    of constant width. A separate histogram is defined for each time,
    and the same bins are used for all times.

    When a new value is added that goes out of range of the current bins,
    bins are doubled in size until that value is within the range of the bins.
    """

    def __init__(self, n_times: int, n_bins: int = 300):
        """
        Args:
            n_times: Length of time dimension of each sample.
            n_bins: Number of bins to use for the histogram. The "effective"
                number of bins may be as small as 1/4th this number of bins,
                as there may be bins greater than the max or less than the
                min value with no samples in them, due to the dynamic resizing
                of bins.
        """
        self._n_times = n_times
        self._n_bins = n_bins
        self.bin_edges: np.ndarray | None = None
        self.counts = np.zeros((n_times, n_bins), dtype=np.int64)
        self._epsilon: float = EPSILON

    def add(self, value: torch.Tensor, i_time_start: int = 0):
        """
        Add new values to the histogram.

        Args:
            value: tensor of values of shape (n_times, n_values) to add to the histogram
            i_time_start: index of the first time to add values to
        """
        # add epsilon to ensure all values stay within (and not just equal to)
        # the bin edges, and to avoid the case where vmin == vmax
        vmin = float((torch.min(value) - self._epsilon).cpu().numpy())
        vmax = float((torch.max(value) + self._epsilon).cpu().numpy())

        if self.bin_edges is None:
            self.bin_edges = np.linspace(vmin, vmax, self._n_bins + 1)
        else:
            while vmin < self.bin_edges[0]:
                self._double_size_left()
            while vmax > self.bin_edges[-1]:
                self._double_size_right()

        i_time_end = i_time_start + value.shape[0]
        for i_time in range(i_time_start, i_time_end):
            try:
                self.counts[i_time] += (
                    histogram(
                        value[i_time - i_time_start, :],
                        self.bin_edges[0],
                        self.bin_edges[1] - self.bin_edges[0],
                        self._n_bins,
                    )
                    .cpu()
                    .numpy()
                )
            except RuntimeError as err:
                # ignore samples with NaNs
                logging.error(f"caught exception while computing histogram: {err}")

    def _double_size_left(self):
        """
        Double the sizes of bins, extending the histogram
        to the left (further negative).
        """
        if self.bin_edges is None:
            raise RuntimeError("Cannot double size of bins without bin edges")
        current_range = self.bin_edges[-1] - self.bin_edges[0]
        new_range = 2 * current_range

        new_bin_edges = np.linspace(
            self.bin_edges[-1] - new_range,
            self.bin_edges[-1],
            self._n_bins + 1,
        )
        new_counts = np.zeros((self._n_times, self._n_bins), dtype=np.int64)
        combined_counts = self.counts[:, ::2] + self.counts[:, 1::2]
        new_counts[:, self._n_bins // 2 :] = combined_counts
        self.bin_edges = new_bin_edges
        self.counts = new_counts

    def _double_size_right(self):
        """
        Double the sizes of bins, extending the histogram
        to the right (further positive).
        """
        if self.bin_edges is None:
            raise RuntimeError("Cannot double size of bins without bin edges")
        current_range = self.bin_edges[-1] - self.bin_edges[0]
        new_range = 2 * current_range

        new_bin_edges = np.linspace(
            self.bin_edges[0],
            self.bin_edges[0] + new_range,
            self._n_bins + 1,
        )
        new_counts = np.zeros((self._n_times, self._n_bins), dtype=np.int64)
        combined_counts = self.counts[:, ::2] + self.counts[:, 1::2]
        new_counts[:, : self._n_bins // 2] = combined_counts
        self.bin_edges = new_bin_edges
        self.counts = new_counts


_Histogram = namedtuple("_Histogram", ["counts", "bin_edges"])


class ComparedDynamicHistograms:
    """Wrapper of DynamicHistogram for multiple histograms, two histograms per
    variable plotted on the same axis.
    """

    def __init__(
        self,
        n_bins: int,
        percentiles: list[float] | None = None,
        compute_percentile_frac: bool = False,
    ) -> None:
        self.n_bins = n_bins
        percentiles = [99.9999] if percentiles is None else percentiles
        self.percentiles = [p for p in percentiles]
        self.target_histograms: Mapping[str, DynamicHistogram] | None = None
        self.prediction_histograms: Mapping[str, DynamicHistogram] | None = None
        self._nan_masks: Mapping[str, torch.Tensor] | None = None
        self._time_dim = -2
        self._variables: set[str] = set()
        self._compute_percentile_frac = compute_percentile_frac

    def _check_overlapping_keys(self, target: TensorMapping, prediction: TensorMapping):
        if not self._variables:
            self._variables = set(target.keys()).intersection(prediction.keys())
        if not self._variables:
            raise ValueError(
                "No overlapping keys between target and prediction variables. "
                f"target: {target.keys()}, prediction: {prediction.keys()}"
            )

        current_variables = set(target.keys()).intersection(prediction.keys())
        if current_variables != self._variables:
            raise ValueError(
                "Available comparison variables provided to record_batch differ "
                f"from initial call to record_batch.  initial: {self._variables}, "
                f"current: {current_variables}"
            )

    @torch.no_grad()
    def record_batch(self, target: TensorMapping, prediction: TensorMapping):
        self._check_overlapping_keys(target, prediction)

        target = {k: v for k, v in target.items() if k in self._variables}
        prediction = {k: v for k, v in prediction.items() if k in self._variables}
        if self._nan_masks is None:
            # assume the final two dimensions are the spatial dimensions and the
            # spatial pattern of NaNs is the same for all samples & time, but
            # can vary by variable
            ndim = next(iter(target.values())).ndim
            index = (slice(0, 1),) * (ndim - 2) + (slice(None), slice(None))
            self._nan_masks = {
                k: v[index].isnan() if torch.any(v[index].isnan()) else None
                for k, v in target.items()
                if k in self._variables
            }

        if self.target_histograms is None or self.prediction_histograms is None:
            self.target_histograms = {}
            for k in target:
                self.target_histograms[k] = DynamicHistogram(
                    n_times=1,
                    n_bins=self.n_bins,
                )
            self.prediction_histograms = {}
            for k in prediction:
                self.prediction_histograms[k] = DynamicHistogram(
                    n_times=1,
                    n_bins=self.n_bins,
                )
        for k in target:
            # no matter what data shape is given, combine it all into one histogram
            value = self._remove_nans(k, target[k])
            self.target_histograms[k].add(value)
        for k in prediction:
            value = self._remove_nans(k, prediction[k])
            self.prediction_histograms[k].add(value)

    def _remove_nans(self, key: str, value: torch.Tensor):
        if self._nan_masks is None:
            raise ValueError(
                "record_batch must be called at least once before removing NaNs"
            )
        nan_mask = self._nan_masks[key]
        if nan_mask is not None:
            mask = ~nan_mask.to(value.device).expand(*value.shape)
            value = torch.masked_select(value, mask)
        else:
            value = value.flatten()
        return value.unsqueeze(0)

    def _get_histograms(
        self,
    ) -> dict[str, dict[Literal["target", "prediction"], _Histogram]]:
        if self.target_histograms is None or self.prediction_histograms is None:
            raise ValueError("No data has been added to the histogram")
        return_dict: dict[str, dict[Literal["target", "prediction"], _Histogram]] = (
            collections.defaultdict(dict)
        )
        for k in self.target_histograms:
            counts, bin_edges = trim_zero_bins(
                self.target_histograms[k].counts.squeeze(self._time_dim),
                self.target_histograms[k].bin_edges,
            )
            return_dict[k]["target"] = _Histogram(counts, bin_edges)
        for k in self.prediction_histograms:
            counts, bin_edges = trim_zero_bins(
                self.prediction_histograms[k].counts.squeeze(self._time_dim),
                self.prediction_histograms[k].bin_edges,
            )
            return_dict[k]["prediction"] = _Histogram(counts, bin_edges)
        return return_dict

    def _plot_histogram(
        self, target_histogram: _Histogram | None, prediction_histogram
    ) -> matplotlib.figure.Figure:
        fig, ax = plt.subplots()
        for histogram, label, line_style, color in zip(
            (target_histogram, prediction_histogram),
            ("target", "prediction"),
            ("-", "--"),
            ("C0", "C1"),
        ):
            normalized_counts = _normalize_histogram(
                histogram.counts, histogram.bin_edges
            )
            if histogram is not None:
                ax.step(
                    histogram.bin_edges[:-1],
                    normalized_counts,
                    line_style,
                    where="post",
                    label=label,
                    color=color,
                )
        ax.set(yscale="log")
        ax.legend()
        plt.tight_layout()
        return fig

    def get_wandb(self) -> dict[str, float]:
        return_dict: dict[str, matplotlib.figure.Figure | float] = {}

        for field_name, histograms in self._get_histograms().items():
            target = histograms.get("target")
            prediction = histograms.get("prediction")
            fig = self._plot_histogram(target, prediction)
            return_dict[field_name] = fig
            plt.close(fig)
            if target is not None:
                for p in self.percentiles:
                    return_dict[f"target/{p}th-percentile/{field_name}"] = quantile(
                        target.bin_edges, target.counts, p / 100.0
                    )
            if prediction is not None:
                for p in self.percentiles:
                    return_dict[f"prediction/{p}th-percentile/{field_name}"] = quantile(
                        prediction.bin_edges, prediction.counts, p / 100.0
                    )
                    if self._compute_percentile_frac and target is not None:
                        return_dict[
                            f"prediction_frac_of_target/{p}th-percentile/{field_name}"
                        ] = (
                            return_dict[f"prediction/{p}th-percentile/{field_name}"]
                            / return_dict[f"target/{p}th-percentile/{field_name}"]
                        )

        return return_dict

    def _get_single_dataset(
        self, histograms: Mapping[str, DynamicHistogram]
    ) -> xr.Dataset:
        data = {}
        for var_name, histogram in histograms.items():
            data[var_name] = xr.DataArray(
                histogram.counts[0, :],
                dims=("bin",),
            )
            data[f"{var_name}_bin_edges"] = xr.DataArray(
                histogram.bin_edges,
                dims=("bin_edges",),
            )
        return xr.Dataset(data)

    def get_dataset(self) -> xr.Dataset:
        if self.target_histograms is None or self.prediction_histograms is None:
            raise ValueError("No data has been added to the histogram")
        target_dataset = self._get_single_dataset(self.target_histograms)
        prediction_dataset = self._get_single_dataset(self.prediction_histograms)
        for missing_target_name in set(prediction_dataset.data_vars) - set(
            target_dataset.data_vars
        ):
            if not missing_target_name.endswith("_bin_edges"):
                target_dataset[missing_target_name] = xr.DataArray(
                    np.zeros_like(prediction_dataset[missing_target_name]),
                    dims=("bin",),
                )
                target_dataset[f"{missing_target_name}_bin_edges"] = prediction_dataset[
                    f"{missing_target_name}_bin_edges"
                ]
        for missing_prediction_name in set(target_dataset.data_vars) - set(
            prediction_dataset.data_vars
        ):
            if not missing_prediction_name.endswith("_bin_edges"):
                prediction_dataset[missing_prediction_name] = xr.DataArray(
                    np.zeros_like(target_dataset[missing_prediction_name]),
                    dims=("bin",),
                )
                prediction_dataset[f"{missing_prediction_name}_bin_edges"] = (
                    target_dataset[f"{missing_prediction_name}_bin_edges"]
                )
        ds = xr.concat([target_dataset, prediction_dataset], dim="source")
        ds["source"] = ["target", "prediction"]
        return ds


def _normalize_histogram(counts, bin_edges):
    """
    Normalize histogram counts so that the integral is 1.
    """
    bin_widths = np.diff(bin_edges)
    return counts / np.sum(counts * bin_widths)



import re
from collections.abc import Iterable, Mapping

import torch

from fme.core.typing_ import TensorMapping


def unstack(tensor: torch.Tensor, names: list[str], dim: int = -1) -> TensorMapping:
    """Unstack a 3D variable to a dictionary of 2D variables.

    Args:
        tensor: 3D tensor to unstack, such as output by a Stacker.
        names: List of names in natural order to assign to the unstacked variables.
            Stacker.get_all_level_names can help in retrieving these names when the
            input tensor was stacked via a Stacker.
        dim: Dimension along which to unstack.

    """
    if len(names) != tensor.size(dim):
        raise ValueError(
            f"Received {len(names)} names, but 3D tensor has {tensor.size(-1)} levels."
        )
    if len(names) == 1:
        return {names[0]: tensor.select(dim=dim, index=0)}
    # split the output tensor along the vertical dimension
    tensors = torch.split(tensor, 1, dim=dim)
    return {name: tensor.squeeze(dim=dim) for name, tensor in zip(names, tensors)}


class Stacker:
    """Handles extraction and stacking of data tensors for 3D variables."""

    LEVEL_PATTERN = re.compile(r"_(\d+)$")

    def __init__(
        self,
        prefix_map: Mapping[str, list[str]] | None = None,
    ):
        """
        Args:
            prefix_map: Mapping which defines the correspondence between an arbitrary
                set of "standard" names (e.g., "surface_pressure" or "air_temperature")
                and lists of possible names or prefix variants (e.g., ["PRESsfc", "PS"]
                or ["air_temperature_", "T_"]) found in the data.
        """
        self._prefix_map: Mapping[str, list[str]] | None = prefix_map

    def infer_prefix_map(self, names: Iterable[str]):
        """
        Infer the prefix map from data names.

        Args:
             names: Variable names, such as the keys from a tensor mapping.
        """
        if self.has_prefix_map:
            raise RuntimeError("Prefix map has already been built.")
        prefix_map = {}
        for name in names:
            match = self.LEVEL_PATTERN.search(name)
            if match is None:
                # 2D variable
                prefix_map[name] = [name]
            else:
                # 3D variable, +1 to include "_" in the prefix
                prefix_name = name[: match.start() + 1]
                prefix_map[prefix_name] = [prefix_name]
        self._prefix_map = prefix_map

    @property
    def has_prefix_map(self) -> bool:
        return self._prefix_map is not None

    @property
    def prefix_map(self) -> Mapping[str, list[str]]:
        """Mapping which defines the correspondence between an arbitrary set of
        "standard" names (e.g., "surface_pressure" or "air_temperature") and
        lists of possible names or prefix variants (e.g., ["PRESsfc", "PS"] or
        ["air_temperature_", "T_"]) found in the data.
        """
        if self._prefix_map is None:
            raise RuntimeError(
                "Stacker's prefix map hasn't yet been built. Build it at runtime by "
                "first calling stacker.infer_prefix_map(data) with the data."
            )
        return self._prefix_map

    @property
    def standard_names(self) -> list[str]:
        return list(self.prefix_map.keys())

    def get_all_level_names(self, standard_name: str, data: TensorMapping) -> list[str]:
        """Get the names of all variables in the data that match one of the
        prefixes associated with the given standard name. If the standard name
        corresponds to a 3D variable, returns all vertical level names in their
        natural order.
        """
        if standard_name not in self.standard_names:
            raise ValueError(f"{standard_name} is not a standard name.")
        for prefix_or_name in self.prefix_map[standard_name]:
            if prefix_or_name in data:
                return [prefix_or_name]
            try:
                return self._natural_sort_names(prefix_or_name, data)
            except KeyError:
                pass
        raise KeyError(
            f"No prefix associated with '{standard_name}' was found in data keys."
        )

    def __call__(self, standard_name: str, data: TensorMapping) -> torch.Tensor:
        """Extract the variable corresponding to standard name and return as a
        3D tensor.

        """
        return self._stack_levels_try(standard_name, data)

    def _stack_levels(self, prefix_or_name: str, data: TensorMapping) -> torch.Tensor:
        names = self._natural_sort_names(prefix_or_name, data)
        # stack along the final dimension
        return torch.stack([data[name] for name in names], dim=-1)

    def _stack_levels_try(
        self, standard_name: str, data: TensorMapping
    ) -> torch.Tensor:
        prefixes_or_names = self.prefix_map[standard_name]
        for prefix_or_name in prefixes_or_names:
            if prefix_or_name in data:
                # 2D variable, return as 1-level 3D tensor
                return data[prefix_or_name].unsqueeze(-1)
            try:
                return self._stack_levels(prefix_or_name, data)
            except KeyError:
                pass
        raise KeyError(
            f"Found no matches for any of {prefixes_or_names} "
            f"among the data names {list(data.keys())}."
        )

    def _natural_sort_names(self, prefix: str, data: TensorMapping) -> list[str]:
        names = [field_name for field_name in data if field_name.startswith(prefix)]

        levels = []
        for name in names:
            match = self.LEVEL_PATTERN.search(name)
            if match is None:
                raise ValueError(
                    f"Invalid field name {name}, is a prefix variable "
                    "but does not end in _{number}."
                )
            levels.append(int(match.group(1)))

        for i, level in enumerate(sorted(levels)):
            if i != level:
                raise ValueError(f"Missing level {i} in {prefix} levels {levels}.")

        if len(names) == 0:
            raise KeyError(prefix)

        return sorted(names, key=lambda name: levels[names.index(name)])



# flake8: noqa
# Copied from tag v0.1.0 https://github.com/NVIDIA/modulus-makani/tree/0218658e925a3c88c6513022a33ecc1647735c2e
# SPDX-FileCopyrightText: Copyright (c) 2024 NVIDIA CORPORATION & AFFILIATES. All rights reserved.
# SPDX-License-Identifier: Apache-2.0
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import torch


@torch.jit.script
def _contract_rank(
    xc: torch.Tensor, wc: torch.Tensor, ac: torch.Tensor, bc: torch.Tensor
) -> torch.Tensor:
    # return torch.einsum("bixy,ior,xr,yr->boxy", x, w, a, b)
    # xc = torch.view_as_complex(x)
    # wc = w #torch.view_as_complex(w)
    # ac = torch.view_as_complex(a)
    # bc = torch.view_as_complex(b)
    resc = torch.einsum("bixy,ior,xr,yr->boxy", xc, wc, ac, bc)
    # res = torch.view_as_real(resc)
    return resc


# # Helper routines for FNOs
@torch.jit.script
def compl_mul1d_fwd(ac: torch.Tensor, bc: torch.Tensor) -> torch.Tensor:
    # ac = torch.view_as_complex(a)
    # bc = torch.view_as_complex(b)
    resc = torch.einsum("bix,io->box", ac, bc)
    # res = torch.view_as_real(resc)
    return resc


@torch.jit.script
def compl_muladd1d_fwd(
    ac: torch.Tensor, bc: torch.Tensor, cc: torch.Tensor
) -> torch.Tensor:
    tmpcc = compl_mul1d_fwd(ac, bc)
    # cc = torch.view_as_complex(c)
    return tmpcc + cc


@torch.jit.script
def compl_mul2d_fwd(ac: torch.Tensor, bc: torch.Tensor) -> torch.Tensor:
    # ac = torch.view_as_complex(a)
    # bc = torch.view_as_complex(b)
    resc = torch.einsum("bixy,io->boxy", ac, bc)
    # res = torch.view_as_real(resc)
    return resc


@torch.jit.script
def compl_muladd2d_fwd(
    ac: torch.Tensor, bc: torch.Tensor, cc: torch.Tensor
) -> torch.Tensor:
    tmpcc = compl_mul2d_fwd(ac, bc)
    # cc = torch.view_as_complex(c)
    return tmpcc + cc


@torch.jit.script
def _contract_localconv_fwd(ac: torch.Tensor, bc: torch.Tensor) -> torch.Tensor:
    # ac = torch.view_as_complex(a)
    # bc = torch.view_as_complex(b)
    resc = torch.einsum("bixy,iox->boxy", ac, bc)
    # res = torch.view_as_real(resc)
    return resc


@torch.jit.script
def _contract_blockconv_fwd(ac: torch.Tensor, bc: torch.Tensor) -> torch.Tensor:
    # ac = torch.view_as_complex(a)
    # bc = torch.view_as_complex(b)
    resc = torch.einsum("bim,imn->bin", ac, bc)
    # res = torch.view_as_real(resc)
    return resc


@torch.jit.script
def _contractadd_blockconv_fwd(
    ac: torch.Tensor, bc: torch.Tensor, cc: torch.Tensor
) -> torch.Tensor:
    tmpcc = _contract_blockconv_fwd(ac, bc)
    # cc = torch.view_as_complex(c)
    return tmpcc + cc


# for the experimental layer
@torch.jit.script
def compl_exp_mul2d_fwd(ac: torch.Tensor, bc: torch.Tensor) -> torch.Tensor:
    # ac = torch.view_as_complex(a)
    # bc = torch.view_as_complex(b)
    resc = torch.einsum("bixy,xio->boxy", ac, bc)
    # res = torch.view_as_real(resc)
    return resc


@torch.jit.script
def compl_exp_muladd2d_fwd(
    ac: torch.Tensor, bc: torch.Tensor, cc: torch.Tensor
) -> torch.Tensor:
    tmpcc = compl_exp_mul2d_fwd(ac, bc)
    # cc = torch.view_as_complex(c)
    return tmpcc + cc


@torch.jit.script
def real_mul2d_fwd(a: torch.Tensor, b: torch.Tensor) -> torch.Tensor:
    res = torch.einsum("bixy,io->boxy", a, b)
    return res


@torch.jit.script
def real_muladd2d_fwd(
    a: torch.Tensor, b: torch.Tensor, c: torch.Tensor
) -> torch.Tensor:
    res = real_mul2d_fwd(a, b) + c
    return res


# new contractions set to replace older ones. We use complex


@torch.jit.script
def _contract_diagonal(ac: torch.Tensor, bc: torch.Tensor) -> torch.Tensor:
    # ac = torch.view_as_complex(a)
    # bc = torch.view_as_complex(b)
    resc = torch.einsum("bixy,ioxy->boxy", ac, bc)
    # res = torch.view_as_real(resc)
    return resc


@torch.jit.script
def _contract_dhconv(ac: torch.Tensor, bc: torch.Tensor) -> torch.Tensor:
    # ac = torch.view_as_complex(a)
    # bc = torch.view_as_complex(b)
    resc = torch.einsum("bixy,iox->boxy", ac, bc)
    # res = torch.view_as_real(resc)
    return resc


@torch.jit.script
def _contract_sep_diagonal(ac: torch.Tensor, bc: torch.Tensor) -> torch.Tensor:
    # ac = torch.view_as_complex(a)
    # bc = torch.view_as_complex(b)
    resc = torch.einsum("bixy,ixy->boxy", ac, bc)
    # res = torch.view_as_real(resc)
    return resc


@torch.jit.script
def _contract_sep_dhconv(ac: torch.Tensor, bc: torch.Tensor) -> torch.Tensor:
    # ac = torch.view_as_complex(a)
    # bc = torch.view_as_complex(b)
    resc = torch.einsum("bixy,ix->boxy", ac, bc)
    # res = torch.view_as_real(resc)
    return resc


@torch.jit.script
def _contract_diagonal_real(a: torch.Tensor, b: torch.Tensor) -> torch.Tensor:
    res = torch.einsum("bixys,ioxy->boxys", a, b).contiguous()
    return res


@torch.jit.script
def _contract_dhconv_real(a: torch.Tensor, b: torch.Tensor) -> torch.Tensor:
    res = torch.einsum("bixys,iox->boxys", a, b).contiguous()
    return res


@torch.jit.script
def _contract_sep_diagonal_real(a: torch.Tensor, b: torch.Tensor) -> torch.Tensor:
    res = torch.einsum("bixys,ixy->boxys", a, b).contiguous()
    return res


@torch.jit.script
def _contract_sep_dhconv_real(a: torch.Tensor, b: torch.Tensor) -> torch.Tensor:
    res = torch.einsum("bixys,ix->boxys", a, b).contiguous()
    return res



import numpy as np
import pytest

from fme.core.winds import lon_lat_to_xyz, u_v_to_x_y_z_wind


def test_u_v_to_x_y_z_wind_energy_conservation():
    """
    Test that the energy is conserved when transforming winds.
    """
    # Define a test case
    u = np.random.randn(10, 10)
    v = np.random.randn(10, 10)
    lat = np.random.uniform(-90, 90, size=(10, 10))
    lon = np.random.uniform(-180, 180, size=(10, 10))

    # Convert to x, y, z
    x, y, z = u_v_to_x_y_z_wind(u, v, lat, lon)

    # Compute the energy conservation equation
    final_energy = x**2 + y**2 + z**2
    initial_energy = u**2 + v**2

    # Check that the energy is conserved
    assert np.allclose(final_energy, initial_energy)


def test_u_v_to_x_y_z_wind_is_horizontal():
    """
    Test that the transformed winds are perpendicular to the vertical vector.
    """
    # Define a test case
    u = np.random.randn(10, 10)
    v = np.random.randn(10, 10)
    lat = np.random.uniform(-90, 90, size=(10, 10))
    lon = np.random.uniform(-180, 180, size=(10, 10))

    # Convert to x, y, z
    wx, wy, wz = u_v_to_x_y_z_wind(u, v, lat, lon)
    x, y, z = lon_lat_to_xyz(lon, lat)

    # Compute the dot product
    dot_product = wx * x + wy * y + wz * z

    # Check that the dot product is zero
    assert np.allclose(dot_product, 0)


@pytest.mark.parametrize(
    "u, v, lat, lon, expected_x, expected_y, expected_z",
    [
        pytest.param(0, 1, 0, 0, 0, 0, 1, id="north_wind_at_equator"),
        pytest.param(1, 0, 0, 0, 0, 1, 0, id="east_wind_at_equator_prime_meridian"),
        pytest.param(
            1, 0, 0, 90, -1, 0, 0, id="east_wind_at_equator_90_degrees_east_meridian"
        ),
        pytest.param(0, -1, 0, 0, 0, 0, -1, id="south_wind_at_equator"),
    ],
)
def test_u_v_to_x_y_z_wind_expected_values(
    u, v, lat, lon, expected_x, expected_y, expected_z
):
    """
    Test that the expected values are returned.
    """
    # Convert to x, y, z
    x, y, z = u_v_to_x_y_z_wind(u, v, lat, lon)

    # Check that the expected values are returned
    assert np.allclose(x, expected_x)
    assert np.allclose(y, expected_y)
    assert np.allclose(z, expected_z)



import numpy as np
import pytest
import torch

from fme.ace.models.healpix.healpix_layers import HEALPixPadding
from fme.core.coordinates import (
    DepthCoordinate,
    HEALPixCoordinates,
    HybridSigmaPressureCoordinate,
    LatLonCoordinates,
    e2ghpx,
)
from fme.core.mask_provider import MaskProvider


@pytest.mark.parametrize(
    "first, second",
    [
        (
            HybridSigmaPressureCoordinate(
                ak=torch.tensor([1, 2, 3]), bk=torch.tensor([4, 5, 6])
            ),
            HybridSigmaPressureCoordinate(
                ak=torch.tensor([1, 2, 3]), bk=torch.tensor([4, 5, 6])
            ),
        ),
        (
            LatLonCoordinates(lat=torch.tensor([1, 2, 3]), lon=torch.tensor([4, 5, 6])),
            LatLonCoordinates(lat=torch.tensor([1, 2, 3]), lon=torch.tensor([4, 5, 6])),
        ),
        (
            HEALPixCoordinates(
                face=torch.tensor([1, 2, 3]),
                height=torch.tensor([4, 5, 6]),
                width=torch.tensor([7, 8, 9]),
            ),
            HEALPixCoordinates(
                face=torch.tensor([1, 2, 3]),
                height=torch.tensor([4, 5, 6]),
                width=torch.tensor([7, 8, 9]),
            ),
        ),
        (
            DepthCoordinate(idepth=torch.tensor([1, 2, 3]), mask=torch.tensor([4, 5])),
            DepthCoordinate(idepth=torch.tensor([1, 2, 3]), mask=torch.tensor([4, 5])),
        ),
    ],
)
def test_equality(first, second):
    assert first == second


@pytest.mark.parametrize(
    "first, second",
    [
        (
            HybridSigmaPressureCoordinate(
                ak=torch.tensor([1, 2, 3]), bk=torch.tensor([4, 5, 6])
            ),
            HybridSigmaPressureCoordinate(
                ak=torch.tensor([1, 2, 3]), bk=torch.tensor([5, 6, 7])
            ),
        ),
        (
            LatLonCoordinates(lat=torch.tensor([1, 2, 3]), lon=torch.tensor([4, 5, 6])),
            LatLonCoordinates(lat=torch.tensor([1, 2, 3]), lon=torch.tensor([5, 6, 7])),
        ),
        (
            HEALPixCoordinates(
                face=torch.tensor([1, 2, 3]),
                height=torch.tensor([4, 5, 6]),
                width=torch.tensor([7, 8, 9]),
            ),
            HEALPixCoordinates(
                face=torch.tensor([1, 2, 3]),
                height=torch.tensor([4, 5, 6]),
                width=torch.tensor([8, 9, 10]),
            ),
        ),
        (
            LatLonCoordinates(lat=torch.tensor([1, 2, 3]), lon=torch.tensor([4, 5, 6])),
            HEALPixCoordinates(
                face=torch.tensor([1, 2, 3]),
                height=torch.tensor([4, 5, 6]),
                width=torch.tensor([7, 8, 9]),
            ),
        ),
        (
            DepthCoordinate(idepth=torch.tensor([1, 2, 3]), mask=torch.tensor([4, 5])),
            DepthCoordinate(idepth=torch.tensor([2, 2, 3]), mask=torch.tensor([4, 5])),
        ),
        (
            DepthCoordinate(idepth=torch.tensor([1, 2, 3]), mask=torch.tensor([4, 5])),
            DepthCoordinate(idepth=torch.tensor([1, 2, 3]), mask=torch.tensor([5, 5])),
        ),
    ],
)
def test_inequality(first, second):
    assert first != second


def test_vertical_integral_shape():
    nlat, nlon, nz = 4, 8, 3
    water = torch.rand(nlat, nlon, nz)
    pressure = torch.rand(nlat, nlon)
    ak, bk = torch.arange(nz + 1), torch.arange(nz + 1)
    coords = HybridSigmaPressureCoordinate(ak, bk)
    water_path = coords.vertical_integral(water, pressure)
    assert water_path.shape == (nlat, nlon)


def test_vertical_coordinates_raises_value_error():
    ak, bk = torch.arange(3), torch.arange(4)
    with pytest.raises(ValueError, match="ak and bk must have the same length"):
        HybridSigmaPressureCoordinate(ak, bk)


def test_vertical_coordinates_len():
    ak, bk = torch.arange(3), torch.arange(3)
    coords = HybridSigmaPressureCoordinate(ak, bk)
    assert len(coords) == 3


def test_interface_pressure():
    ak = torch.tensor([2.0, 0.5, 0.0])
    bk = torch.tensor([0.0, 0.5, 1.0])
    psfc = torch.tensor([[1, 1], [2, 2]])
    coords = HybridSigmaPressureCoordinate(ak, bk)
    pinterface = coords.interface_pressure(psfc)
    assert pinterface.shape == (2, 2, 3)
    assert pinterface[0, 0, 0] == ak[0]
    assert pinterface[0, 0, -1] == bk[-1] * psfc[0, 0]


@pytest.mark.parametrize(
    "idepth, mask, msg",
    [
        (torch.tensor([[1, 2, 3]]), torch.tensor([4, 5]), "1-dimensional tensor"),
        (torch.tensor([1, 2, 3]), torch.tensor([4, 5, 6]), "last dimension of mask"),
        (torch.tensor([1]), torch.tensor([]), "idepth must have at least two"),
    ],
    ids=["idepth 2D", "mask inconsistent with idepth", "idepth too short"],
)
def test_depth_coordinate_validation(idepth, mask, msg):
    with pytest.raises(ValueError, match=msg):
        DepthCoordinate(idepth, mask)


def test_depth_coordinate_integral_raises():
    idepth = torch.tensor([1, 2, 3])
    mask = torch.tensor([1, 1])
    data = torch.tensor([1, 1, 4])
    coords = DepthCoordinate(idepth, mask)
    with pytest.raises(ValueError, match="dimension of integrand must match"):
        coords.depth_integral(data)


@pytest.mark.parametrize(
    "idepth, mask, expected",
    [
        (torch.tensor([1, 2, 3]), torch.tensor([1, 1]), torch.tensor(3.0)),
        (torch.tensor([1, 2, 3]), torch.tensor([1, 0]), torch.tensor(1.0)),
        (torch.tensor([1, 2, 3]), torch.tensor([0, 0]), torch.tensor(float("nan"))),
        (torch.tensor([1, 2, 4]), torch.tensor([1, 1]), torch.tensor(5.0)),
    ],
    ids=[
        "mask is all ones",
        "second layer is masked out",
        "all layers masked out",
        "varying depth",
    ],
)
def test_depth_integral_1d_data(idepth, mask, expected):
    data = torch.arange(1, len(idepth))
    result = DepthCoordinate(idepth, mask).depth_integral(data)
    torch.testing.assert_close(result, expected, equal_nan=True)


def test_depth_integral_3d_data():
    img_shape = 2, 4
    idepth = torch.tensor([1, 2, 4])
    mask = torch.tensor([1, 1])
    nz = len(mask)
    data = torch.arange(1, img_shape[0] * img_shape[1] * nz + 1).reshape(
        img_shape + (nz,)
    )
    depth_0 = idepth[1] - idepth[0]
    depth_1 = idepth[2] - idepth[1]
    expected = (data[:, :, 0] * depth_0 + data[:, :, 1] * depth_1).float()
    result = DepthCoordinate(idepth, mask).depth_integral(data)
    torch.testing.assert_close(result, expected)


@pytest.mark.parametrize(
    "name, level",
    [
        ("sfc_level", 0),
        ("depth_0", 0),
        ("depth_3", 3),
    ],
)
def test_depth_get_mask_tensor_for(name, level):
    idepth = torch.arange(end=5)
    mask = torch.arange(end=4)
    coord = DepthCoordinate(idepth, mask)
    assert coord.get_mask_tensor_for(name) == level


def test_depth_returns_surface_mask_if_specified():
    idepth = torch.arange(end=5)
    mask = torch.arange(end=4)
    surface_mask = torch.tensor([4])
    coord_sfc_mask = DepthCoordinate(idepth, mask, surface_mask)
    coord_no_sfc_mask = DepthCoordinate(idepth, mask)
    assert coord_sfc_mask.get_mask_tensor_for("sfc_level") == surface_mask[0]
    assert coord_no_sfc_mask.get_mask_tensor_for("sfc_level") == mask[0]


def test_masked_lat_lon_ops_from_coords():
    lat = torch.tensor([0.0, 0.0, 0.0])
    lon = torch.tensor([0.0])
    mask = torch.tensor([[1], [0], [1]])
    coords = LatLonCoordinates(lat=lat, lon=lon)
    mask_provider = MaskProvider(masks={"mask_0": mask})
    gridded_ops = coords.get_gridded_operations(mask_provider=mask_provider)
    input_ = torch.tensor([[1.0], [-10.0], [3.0]])
    result = gridded_ops.area_weighted_mean(input_, name="T_0")
    torch.testing.assert_close(result, torch.tensor(2.0))


def test_healpix_ops_raises_value_error_with_mask():
    face = torch.arange(12)
    height = torch.arange(16)
    width = torch.arange(16)
    healpix_coords = HEALPixCoordinates(face=face, height=height, width=width)
    mask_provider = MaskProvider(masks={"mask_0": torch.tensor([1, 0, 1])})

    expected_msg = "HEALPixCoordinates does not support a mask"
    with pytest.raises(NotImplementedError, match=expected_msg):
        healpix_coords.get_gridded_operations(mask_provider=mask_provider)


@pytest.mark.skipif(e2ghpx is None, reason="earth2grid healpix not available")
@pytest.mark.parametrize("pad", [True, False])
def test_healpix_coordinates_xyz(pad: bool, very_fast_only: bool):
    if very_fast_only:
        pytest.skip("Skipping non-fast tests and healpix (earth2grid) tests")

    face = torch.arange(12)
    height = torch.arange(16)
    width = torch.arange(16)
    healpix_coords = HEALPixCoordinates(face=face, height=height, width=width)

    x, y, z = healpix_coords.xyz

    # Calculate original distances between points along x and y axes
    distances_x = np.sqrt(
        (np.diff(x, axis=-1) ** 2)
        + (np.diff(y, axis=-1) ** 2)
        + (np.diff(z, axis=-1) ** 2)
    )
    distances_y = np.sqrt(
        (np.diff(x, axis=-2) ** 2)
        + (np.diff(y, axis=-2) ** 2)
        + (np.diff(z, axis=-2) ** 2)
    )

    # Apply HEALPix padding
    if pad:
        padding = 2
        healpix_padding = HEALPixPadding(padding=padding, enable_nhwc=False)
        padded_x = healpix_padding(torch.Tensor(x).unsqueeze(1)).squeeze(1)
        padded_y = healpix_padding(torch.Tensor(y).unsqueeze(1)).squeeze(1)
        padded_z = healpix_padding(torch.Tensor(z).unsqueeze(1)).squeeze(1)

        # Calculate distances between padding points along x and y axes
        distances_padded_x = np.sqrt(
            (np.diff(padded_x.numpy(), axis=-1) ** 2)
            + (np.diff(padded_y.numpy(), axis=-1) ** 2)
            + (np.diff(padded_z.numpy(), axis=-1) ** 2)
        )
        distances_padded_y = np.sqrt(
            (np.diff(padded_x.numpy(), axis=-2) ** 2)
            + (np.diff(padded_y.numpy(), axis=-2) ** 2)
            + (np.diff(padded_z.numpy(), axis=-2) ** 2)
        )

        max_distance_x = np.max(distances_x)
        max_distance_y = np.max(distances_y)

        # Assert distances are not too far from the expected diff (min = 0, max = 2dx)
        assert np.all(
            distances_padded_x <= 2 * max_distance_x
        ), "Some distances along x-axis exceed 2 * max_distance_x"
        assert np.all(
            distances_padded_y <= 2 * max_distance_y
        ), "Some distances along y-axis exceed 2 * max_distance_y"

    else:
        mean_distances_x = distances_x.mean()
        mean_distances_y = distances_y.mean()

        assert np.allclose(distances_x, mean_distances_x, atol=0.03)
        assert np.allclose(distances_y, mean_distances_y, atol=0.03)



import pytest

from fme.core.typing_ import Slice


@pytest.mark.parametrize(
    "start, stop, step, value, expected",
    [
        (0, 10, 1, 0, True),
        (0, 10, 1, 10, False),
        (0, 10, 2, 0, True),
        (0, 10, 2, 1, False),
        (None, None, None, 0, True),
        (None, None, None, 500, True),
        (0, 10, None, 4, True),
        (0, 10, None, 10, False),
        (0, None, 2, 0, True),
        (None, None, 2, 1, False),
    ],
)
def test_slice_contains(start, stop, step, value, expected):
    slice = Slice(start=start, stop=stop, step=step)
    assert slice.contains(value) == expected



import dataclasses
import logging
from collections.abc import Callable, Mapping
from typing import Protocol

import numpy as np
import torch
import xarray as xr

from fme.core.dataset_info import DatasetInfo
from fme.core.diagnostics import get_reduced_diagnostics, write_reduced_diagnostics
from fme.core.generics.aggregator import AggregatorABC
from fme.core.typing_ import TensorDict, TensorMapping

from .map import MapAggregator
from .reduced import MeanAggregator
from .snapshot import SnapshotAggregator
from .spectrum import SpectrumAggregator


class _Aggregator(Protocol):
    def get_logs(self, label: str) -> TensorMapping: ...

    def record_batch(
        self,
        loss: float,
        target_data: TensorMapping,
        gen_data: TensorMapping,
        target_data_norm: TensorMapping,
        gen_data_norm: TensorMapping,
    ) -> None: ...

    def get_dataset(self) -> xr.Dataset: ...


@dataclasses.dataclass
class DeterministicTrainOutput:
    metrics: TensorDict
    gen_data: TensorDict
    target_data: TensorDict
    normalize: Callable[[TensorDict], TensorDict]


class OneStepDeterministicAggregator(AggregatorABC[DeterministicTrainOutput]):
    """
    Aggregates statistics for the first timestep.

    Operates on batches without an ensemble dimension (i.e. shape [batch, ...]).

    To use, call `record_batch` on the results of each batch, then call
    `get_logs` to get a dictionary of statistics when you're done.
    """

    def __init__(
        self,
        dataset_info: DatasetInfo,
        save_diagnostics: bool = True,
        output_dir: str | None = None,
        loss_scaling: TensorMapping | None = None,
        log_snapshots: bool = True,
        log_mean_maps: bool = True,
    ):
        """
        Args:
            dataset_info: Dataset coordinates and metadata.
            save_diagnostics: Whether to save diagnostics.
            output_dir: Directory to write diagnostics to.
            variable_metadata: Metadata for each variable.
            loss_scaling: Dictionary of variables and their scaling factors
                used in loss computation.
            log_snapshots: Whether to include snapshots in diagnostics.
            log_mean_maps: Whether to include mean maps in diagnostics.
        """
        if save_diagnostics and output_dir is None:
            raise ValueError("Output directory must be set to save diagnostics.")
        self._output_dir = output_dir
        self._save_diagnostics = save_diagnostics
        horizontal_coordinates = dataset_info.horizontal_coordinates
        self._coords = horizontal_coordinates.coords
        self._aggregators: dict[str, _Aggregator] = {
            "mean": MeanAggregator(dataset_info.gridded_operations),
        }
        try:
            self._aggregators["power_spectrum"] = SpectrumAggregator(
                dataset_info.gridded_operations,
            )
        except NotImplementedError:
            logging.warning(
                "Spectrum aggregator not implemented for this grid type, omitting."
            )
        if log_snapshots:
            self._aggregators["snapshot"] = SnapshotAggregator(
                horizontal_coordinates.dims, dataset_info.variable_metadata
            )
        if log_mean_maps:
            self._aggregators["mean_map"] = MapAggregator(
                horizontal_coordinates.dims, dataset_info.variable_metadata
            )

        self._loss_scaling = loss_scaling or {}

    @torch.no_grad()
    def record_batch(
        self,
        batch: DeterministicTrainOutput,
    ):
        if len(batch.target_data) == 0:
            raise ValueError("No data in target_data")
        if len(batch.gen_data) == 0:
            raise ValueError("No data in gen_data")

        target_data_norm = batch.normalize(batch.target_data)
        gen_data_norm = batch.normalize(batch.gen_data)
        for agg in self._aggregators.values():
            agg.record_batch(
                loss=batch.metrics.get("loss", np.nan),
                target_data=batch.target_data,
                gen_data=batch.gen_data,
                target_data_norm=target_data_norm,
                gen_data_norm=gen_data_norm,
            )

    @torch.no_grad()
    def get_logs(self, label: str):
        """
        Returns logs as can be reported to WandB.

        Args:
            label: Label to prepend to all log keys.
        """
        logs = {}
        for agg_label in self._aggregators:
            logging.info(f"Getting logs for {agg_label} aggregator")
            for k, v in self._aggregators[agg_label].get_logs(label=agg_label).items():
                logs[f"{label}/{k}"] = v
        logging.info(f"Inserting loss-scaled MSE componenets into logs")
        logs.update(
            self._get_loss_scaled_mse_components(
                validation_metrics=logs,
                label=label,
            )
        )
        return logs

    def _get_loss_scaled_mse_components(
        self,
        validation_metrics: Mapping[str, float],
        label: str,
    ):
        scaled_squared_errors = {}

        for var in self._loss_scaling:
            rmse_key = f"{label}/mean/weighted_rmse/{var}"
            if rmse_key in validation_metrics:
                scaled_squared_errors[var] = (
                    validation_metrics[rmse_key] / self._loss_scaling[var].item()
                ) ** 2
        scaled_squared_errors_sum = sum(scaled_squared_errors.values())
        fractional_contribs = {
            f"{label}/mean/mse_fractional_components/{k}": v / scaled_squared_errors_sum
            for k, v in scaled_squared_errors.items()
        }
        return fractional_contribs

    @torch.no_grad()
    def flush_diagnostics(self, subdir: str | None = None):
        if self._save_diagnostics:
            reduced_diagnostics = get_reduced_diagnostics(
                sub_aggregators=self._aggregators,
                coords=self._coords,
            )
            if self._output_dir is not None:
                write_reduced_diagnostics(
                    reduced_diagnostics,
                    self._output_dir,
                    subdir=subdir,
                )
            else:
                raise ValueError("Output directory is not set.")



import logging
import random
from collections.abc import Callable, Iterator

import torch

from fme.ace.data_loading.batch_data import BatchData
from fme.core.dataset.xarray import XarrayDataset


def get_data_loader(
    dataset: torch.utils.data.Dataset,
    batch_size: int,
    n_window_timesteps: int,
    time_buffer: int,
    num_workers: int,
    sampler: torch.utils.data.Sampler,
    shuffled: bool,
    drop_last: bool,
    pin_memory: bool,
    collate_fn: Callable,
    multiprocessing_context: str | None,
    persistent_workers: bool,
    prefetch_factor: int | None,
) -> torch.utils.data.DataLoader[BatchData] | "SlidingWindowDataLoader":
    if prefetch_factor is None:
        # DataLoader default is not None so we must leave it unset
        kwargs = {}
    else:
        kwargs = {"prefetch_factor": prefetch_factor}
    dataloader = torch.utils.data.DataLoader(
        dataset,
        batch_size=batch_size,
        num_workers=num_workers,
        sampler=sampler,
        drop_last=drop_last,
        pin_memory=pin_memory,
        collate_fn=collate_fn,
        multiprocessing_context=multiprocessing_context,
        persistent_workers=persistent_workers,
        **kwargs,
    )
    if time_buffer > 0:
        n_timesteps_preloaded = time_buffer + n_window_timesteps
        dataloader = SlidingWindowDataLoader(
            dataloader,
            n_timesteps_preloaded,
            n_window_timesteps,
            shuffled,
            dataset=dataloader.dataset,
        )

    if len(dataloader) == 0:
        msg = (
            "No batches in dataloader: "
            f"{len(dataloader.dataset)} samples, "
            f"batch size is {dataloader.batch_size}"
        )
        if time_buffer > 0:
            msg += f", and an outer sample length is {n_timesteps_preloaded}"
        raise ValueError(msg)

    return dataloader


class SlidingWindowDataLoader:
    """
    A wrapper for the torch data loader that provides additional properties.
    Create more batches by applying a sliding window to existing batches.
    """

    def __init__(
        self,
        loader: torch.utils.data.DataLoader[BatchData],
        input_n_timesteps: int,
        output_n_timesteps: int,
        shuffle: bool,
        dataset: XarrayDataset,
    ):
        """
        Create a sliding window over the input data loader to generate new batches.
            We assume that the loader returns batches with the same number of timesteps
            equal to "input_n_timesteps". This is asserted for every batch in __next__.

        Args:
            loader: Pre-batched input data loader that returns batches of
                `input_n_timesteps` timesteps.
            input_n_timesteps: Number of timesteps in the input batch, i.e., the size of
                the sliding window.
            output_n_timesteps: Number of timesteps in the output batch, must be smaller
                than the number of input timesteps.
            shuffle: Whether to shuffle the start indices for the sliding window.
            dataset: The underlying dataset associated with the loader.
        """
        self._input_loader_len = len(loader)
        self._loader = loader
        assert output_n_timesteps <= input_n_timesteps
        self._input_n_timesteps = input_n_timesteps
        self._output_n_timesteps = output_n_timesteps
        self._n_new_batches = input_n_timesteps - output_n_timesteps + 1
        self._shuffle = shuffle
        self.dataset = dataset

    def __iter__(self) -> Iterator[BatchData]:
        # reset the iterator state
        self._loaditer = iter(self._loader)
        self._current_batch: BatchData | None = None
        return self

    def __len__(self) -> int:
        return self._input_loader_len * self._n_new_batches

    def __next__(self) -> BatchData:
        if self._current_batch is None:
            self._current_batch = next(self._loaditer)
            self._counter = 0
            assert self._current_batch.n_timesteps == self._input_n_timesteps
            self._start_steps = list(range(0, self._n_new_batches))
            if self._shuffle:
                random.shuffle(self._start_steps)
        start = self._start_steps[self._counter]
        slice_ = slice(start, start + self._output_n_timesteps)
        small_batch = self._current_batch.select_time_slice(slice_)
        self._counter += 1
        if self._counter == self._n_new_batches:
            self._current_batch = None
        return small_batch

    @property
    def batch_size(self) -> int:
        """
        Number of samples in a single batch.
        """
        return self._loader.batch_size

    def _n_samples_per_dataset_item(self) -> int:
        """
        Number of samples per dataset outer item, i.e., in a window.
        """
        return self._input_n_timesteps - self._output_n_timesteps + 1

    def log_info(self, name: str):
        logging.info(
            f"{name} data: {len(self) * self.batch_size} samples, {len(self)} batches"
        )
        logging.info(
            f"{name} data: An outer window size of "
            f"{self._n_samples_per_dataset_item()} samples was used to load "
            "the above samples and batches in chunks."
        )



# flake8: noqa
# Copied from https://github.com/ai2cm/modulus/commit/22df4a9427f5f12ff6ac891083220e7f2f54d229
# Copyright (c) 2023, NVIDIA CORPORATION & AFFILIATES. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import dataclasses
import math
from typing import Tuple

import torch
import torch.fft
import torch.nn as nn
import torch.nn.functional as F
from torch.utils.checkpoint import checkpoint

from .activations import ComplexReLU
from .contractions import compl_mul2d_fwd, compl_muladd2d_fwd


@dataclasses.dataclass
class ContextConfig:
    """
    Configuration for the context.
    """

    embed_dim_scalar: int
    embed_dim_2d: int


@dataclasses.dataclass
class Context:
    """
    Context for the conditional layer normalization.

    Parameters:
        embedding_scalar: The scalar embedding to condition on. The
            last dimension is the channel dimension.
        embedding_2d: The 2D embedding to condition on. The last
            three dimensions are (channels, height, width).
    """

    embedding_scalar: torch.Tensor | None
    embedding_2d: torch.Tensor | None

    def __post_init__(self):
        if (
            self.embedding_scalar is not None
            and self.embedding_2d is not None
            and self.embedding_2d.ndim != self.embedding_scalar.ndim + 2
        ):
            raise ValueError(
                "embedding_2d must have 2 more dimensions than embedding_scalar"
            )


class ConditionalLayerNorm(nn.Module):
    """
    Conditional Layer Normalization as described in "AdaSpeech: Adaptive
    Text to Speech for Custom Voice" https://arxiv.org/abs/2103.00993.

    Assumes that the input has shape (batch_size, channels, height, width).
    """

    def __init__(
        self,
        n_channels: int,
        img_shape: Tuple[int, int],
        context_config: ContextConfig,
        global_layer_norm: bool = False,
        epsilon: float = 1e-5,
    ):
        super(ConditionalLayerNorm, self).__init__()
        self.n_channels = n_channels
        self.embed_dim_scalar = context_config.embed_dim_scalar
        self.embed_dim_2d = context_config.embed_dim_2d
        self.epsilon = epsilon
        if self.embed_dim_scalar > 0:
            self.W_scale: nn.Linear | None = nn.Linear(
                self.embed_dim_scalar, self.n_channels
            )
            self.W_bias: nn.Linear | None = nn.Linear(
                self.embed_dim_scalar, self.n_channels
            )
        else:
            self.W_scale = None
            self.W_bias = None
        if self.embed_dim_2d > 0:
            # no bias as it is already handled in the non-2d layers
            self.W_scale_2d = nn.Conv2d(
                self.embed_dim_2d, self.n_channels, kernel_size=1, bias=False
            )
            self.W_bias_2d = nn.Conv2d(
                self.embed_dim_2d, self.n_channels, kernel_size=1, bias=False
            )
        else:
            self.W_scale_2d = None
            self.W_bias_2d = None
        if global_layer_norm:
            self.norm = nn.LayerNorm(
                (self.n_channels, img_shape[0], img_shape[1]),
                eps=epsilon,
                elementwise_affine=False,
            )
        else:
            self.norm = nn.LayerNorm(
                (self.n_channels,),
                eps=epsilon,
                elementwise_affine=False,
            )
        self._global_layer_norm = global_layer_norm
        self.reset_parameters()

    def reset_parameters(self):
        if self.W_scale is not None:
            torch.nn.init.constant_(self.W_scale.weight, 0.0)
            torch.nn.init.constant_(self.W_scale.bias, 1.0)
        if self.W_bias is not None:
            torch.nn.init.constant_(self.W_bias.weight, 0.0)
            torch.nn.init.constant_(self.W_bias.bias, 0.0)
        if self.W_scale_2d is not None:
            torch.nn.init.constant_(self.W_scale_2d.weight, 0.0)
        if self.W_bias_2d is not None:
            torch.nn.init.constant_(self.W_bias_2d.weight, 0.0)
        # no bias on 2d layers as it is already handled in the non-2d layers

    def forward(self, x: torch.Tensor, context: Context) -> torch.Tensor:
        """
        Conditional Layer Normalization

        This is a modified version of LayerNorm that allows the scale and bias to be
        conditioned on a context embedding.

        Args:
            x: The input tensor to normalize, of shape
                (batch_size, channels, height, width).
            context: The context to condition on.

        Returns:
            The normalized tensor, of shape (batch_size, channels, height, width).
        """
        if self.W_scale is not None:
            if context.embedding_scalar is None:
                raise ValueError("embedding_scalar must be provided")
            scale: torch.Tensor = (
                self.W_scale(context.embedding_scalar).unsqueeze(-1).unsqueeze(-1)
            )
        else:
            scale = torch.ones_like(x)
        if self.W_scale_2d is not None:
            if context.embedding_2d is None:
                raise ValueError("embedding_2d must be provided")
            scale = scale + self.W_scale_2d(context.embedding_2d)
        if self.W_bias is not None:
            if context.embedding_scalar is None:
                raise ValueError("embedding_scalar must be provided")
            bias: torch.Tensor = (
                self.W_bias(context.embedding_scalar).unsqueeze(-1).unsqueeze(-1)
            )
        else:
            bias = torch.zeros_like(x)
        if self.W_bias_2d is not None:
            if context.embedding_2d is None:
                raise ValueError("embedding_2d must be provided")
            bias = bias + self.W_bias_2d(context.embedding_2d)
        if not self._global_layer_norm:
            x = x.transpose(1, -1)
        x_norm: torch.Tensor = self.norm(x)
        if not self._global_layer_norm:
            x_norm = x_norm.transpose(1, -1)
        return x_norm * scale + bias


@torch.jit.script
def drop_path(
    x: torch.Tensor, drop_prob: float = 0.0, training: bool = False
) -> torch.Tensor:  # pragma: no cover
    """Drop paths (Stochastic Depth) per sample (when applied in main path of
    residual blocks).
    This is the same as the DropConnect impl for EfficientNet, etc networks, however,
    the original name is misleading as 'Drop Connect' is a different form of dropout in
    a separate paper. See discussion:
        https://github.com/tensorflow/tpu/issues/494#issuecomment-532968956
    We've opted for changing the layer and argument names to 'drop path' rather than
    mix DropConnect as a layer name and use 'survival rate' as the argument.
    """
    if drop_prob == 0.0 or not training:
        return x
    keep_prob = 1.0 - drop_prob
    shape = (x.shape[0],) + (1,) * (
        x.ndim - 1
    )  # work with diff dim tensors, not just 2d ConvNets
    random_tensor = keep_prob + torch.rand(shape, dtype=x.dtype, device=x.device)
    random_tensor.floor_()  # binarize
    output = x.div(keep_prob) * random_tensor
    return output


class DropPath(nn.Module):
    """
    Drop paths (Stochastic Depth) per sample (when applied in main path of residual
    blocks).
    """

    def __init__(self, drop_prob=None):  # pragma: no cover
        super(DropPath, self).__init__()
        self.drop_prob = drop_prob

    def forward(self, x):  # pragma: no cover
        return drop_path(x, self.drop_prob, self.training)


class PatchEmbed(nn.Module):
    """
    Divides the input image into patches and embeds them into a specified dimension
    using a convolutional layer.
    """

    def __init__(
        self, img_size=(224, 224), patch_size=(16, 16), in_chans=3, embed_dim=768
    ):  # pragma: no cover
        super(PatchEmbed, self).__init__()
        num_patches = (img_size[1] // patch_size[1]) * (img_size[0] // patch_size[0])
        self.img_size = img_size
        self.patch_size = patch_size
        self.num_patches = num_patches
        self.proj = nn.Conv2d(
            in_chans, embed_dim, kernel_size=patch_size, stride=patch_size
        )

    def forward(self, x):  # pragma: no cover
        # gather input
        B, C, H, W = x.shape
        assert (
            H == self.img_size[0] and W == self.img_size[1]
        ), f"Input image size ({H}*{W}) doesn't match model ({self.img_size[0]}*{self.img_size[1]})."
        # new: B, C, H*W
        x = self.proj(x).flatten(2)
        return x


class MLP(nn.Module):
    """
    Basic CNN with support for gradient checkpointing
    """

    def __init__(
        self,
        in_features,
        hidden_features=None,
        out_features=None,
        act_layer=nn.GELU,
        output_bias=True,
        drop_rate=0.0,
        checkpointing=0,
    ):  # pragma: no cover
        super(MLP, self).__init__()
        self.checkpointing = checkpointing
        out_features = out_features or in_features
        hidden_features = hidden_features or in_features

        fc1 = nn.Conv2d(in_features, hidden_features, 1, bias=True)
        act = act_layer()
        fc2 = nn.Conv2d(hidden_features, out_features, 1, bias=output_bias)
        if drop_rate > 0.0:
            drop = nn.Dropout(drop_rate)
            self.fwd = nn.Sequential(fc1, act, drop, fc2, drop)
        else:
            self.fwd = nn.Sequential(fc1, act, fc2)

        # by default, all weights are shared

    @torch.jit.ignore
    def checkpoint_forward(self, x):  # pragma: no cover
        """Forward method with support for gradient checkpointing"""
        return checkpoint(self.fwd, x)

    def forward(self, x):  # pragma: no cover
        if self.checkpointing >= 2:
            return self.checkpoint_forward(x)
        else:
            return self.fwd(x)


class RealFFT2(nn.Module):
    """
    Helper routine to wrap FFT similarly to the SHT
    """

    def __init__(self, nlat, nlon, lmax=None, mmax=None):  # pragma: no cover
        super(RealFFT2, self).__init__()

        # use local FFT here
        self.fft_handle = torch.fft.rfft2

        self.nlat = nlat
        self.nlon = nlon
        self.lmax = lmax or self.nlat
        self.mmax = mmax or self.nlon // 2 + 1

        self.truncate = True
        if (self.lmax == self.nlat) and (self.mmax == (self.nlon // 2 + 1)):
            self.truncate = False

        # self.num_batches = 1
        assert self.lmax % 2 == 0

    def forward(self, x):  # pragma: no cover
        y = self.fft_handle(x, (self.nlat, self.nlon), (-2, -1), "ortho")

        if self.truncate:
            y = torch.cat(
                (
                    y[..., : math.ceil(self.lmax / 2), : self.mmax],
                    y[..., -math.floor(self.lmax / 2) :, : self.mmax],
                ),
                dim=-2,
            )

        return y


class InverseRealFFT2(nn.Module):
    """
    Helper routine to wrap FFT similarly to the SHT
    """

    def __init__(self, nlat, nlon, lmax=None, mmax=None):  # pragma: no cover
        super(InverseRealFFT2, self).__init__()

        # use local FFT here
        self.ifft_handle = torch.fft.irfft2

        self.nlat = nlat
        self.nlon = nlon
        self.lmax = lmax or self.nlat
        self.mmax = mmax or self.nlon // 2 + 1

    def forward(self, x):  # pragma: no cover
        out = self.ifft_handle(x, (self.nlat, self.nlon), (-2, -1), "ortho")

        return out


class SpectralAttention2d(nn.Module):
    """
    2d Spectral Attention layer
    """

    def __init__(
        self,
        forward_transform,
        inverse_transform,
        embed_dim,
        sparsity_threshold=0.0,
        hidden_size_factor=2,
        use_complex_network=True,
        use_complex_kernels=False,
        complex_activation="real",
        bias=False,
        spectral_layers=1,
        drop_rate=0.0,
    ):  # pragma: no cover
        super(SpectralAttention2d, self).__init__()

        self.embed_dim = embed_dim
        self.sparsity_threshold = sparsity_threshold
        self.hidden_size = int(hidden_size_factor * self.embed_dim)
        self.scale = 0.02
        self.spectral_layers = spectral_layers
        if use_complex_kernels:
            raise NotImplementedError("complex kernels not supported")
        self.mul_add_handle = compl_muladd2d_fwd
        self.mul_handle = compl_mul2d_fwd

        self.modes_lat = forward_transform.lmax
        self.modes_lon = forward_transform.mmax

        # only storing the forward handle to be able to call it
        self.forward_transform = forward_transform.forward
        self.inverse_transform = inverse_transform.forward

        assert inverse_transform.lmax == self.modes_lat
        assert inverse_transform.mmax == self.modes_lon

        # weights
        w = [self.scale * torch.randn(self.embed_dim, self.hidden_size, 2)]
        # w = [self.scale * torch.randn(self.embed_dim + 2*self.embed_freqs, self.hidden_size, 2)]
        # w = [self.scale * torch.randn(self.embed_dim + 4*self.embed_freqs, self.hidden_size, 2)]
        for l in range(1, self.spectral_layers):
            w.append(self.scale * torch.randn(self.hidden_size, self.hidden_size, 2))
        self.w = nn.ParameterList(w)

        if bias:
            self.b = nn.ParameterList(
                [
                    self.scale * torch.randn(self.hidden_size, 1, 2)
                    for _ in range(self.spectral_layers)
                ]
            )

        self.wout = nn.Parameter(
            self.scale * torch.randn(self.hidden_size, self.embed_dim, 2)
        )

        self.drop = nn.Dropout(drop_rate) if drop_rate > 0.0 else nn.Identity()

        self.activation = ComplexReLU(
            mode=complex_activation, bias_shape=(self.hidden_size, 1, 1)
        )

    def forward_mlp(self, xr):  # pragma: no cover
        """forward method for the MLP part of the network"""
        for l in range(self.spectral_layers):
            if hasattr(self, "b"):
                xr = self.mul_add_handle(
                    xr, self.w[l].to(xr.dtype), self.b[l].to(xr.dtype)
                )
            else:
                xr = self.mul_handle(xr, self.w[l].to(xr.dtype))
            xr = torch.view_as_complex(xr)
            xr = self.activation(xr)
            xr = self.drop(xr)
            xr = torch.view_as_real(xr)

        xr = self.mul_handle(xr, self.wout)

        return xr

    def forward(self, x):  # pragma: no cover
        dtype = x.dtype
        # x = x.to(torch.float32)

        # FWD transform
        with torch.amp.autocast("cuda", enabled=False):
            x = x.to(torch.float32)
            x = x.contiguous()
            x = self.forward_transform(x)
            x = torch.view_as_real(x)

        # MLP
        x = self.forward_mlp(x)

        # BWD transform
        with torch.amp.autocast("cuda", enabled=False):
            x = torch.view_as_complex(x)
            x = x.contiguous()
            x = self.inverse_transform(x)
            x = x.to(dtype)

        return x


class SpectralAttentionS2(nn.Module):
    """
    geometrical Spectral Attention layer
    """

    def __init__(
        self,
        forward_transform,
        inverse_transform,
        embed_dim,
        sparsity_threshold=0.0,
        hidden_size_factor=2,
        use_complex_network=True,
        use_complex_kernels=False,
        complex_activation="real",
        bias=False,
        spectral_layers=1,
        drop_rate=0.0,
    ):  # pragma: no cover
        super(SpectralAttentionS2, self).__init__()

        self.embed_dim = embed_dim
        self.sparsity_threshold = sparsity_threshold
        self.hidden_size = int(hidden_size_factor * self.embed_dim)
        self.scale = 0.02
        if use_complex_kernels:
            raise NotImplementedError("complex kernels not supported")
        # self.mul_add_handle = compl_muladd1d_fwd_c if use_complex_kernels else compl_muladd1d_fwd
        self.mul_add_handle = compl_muladd2d_fwd
        # self.mul_handle = compl_mul1d_fwd_c if use_complex_kernels else compl_mul1d_fwd
        self.mul_handle = compl_mul2d_fwd
        self.spectral_layers = spectral_layers

        self.modes_lat = forward_transform.lmax
        self.modes_lon = forward_transform.mmax

        # only storing the forward handle to be able to call it
        self.forward_transform = forward_transform.forward
        self.inverse_transform = inverse_transform.forward

        assert inverse_transform.lmax == self.modes_lat
        assert inverse_transform.mmax == self.modes_lon

        # weights
        w = [self.scale * torch.randn(self.embed_dim, self.hidden_size, 2)]
        # w = [self.scale * torch.randn(self.embed_dim + 4*self.embed_freqs, self.hidden_size, 2)]
        for l in range(1, self.spectral_layers):
            w.append(self.scale * torch.randn(self.hidden_size, self.hidden_size, 2))
        self.w = nn.ParameterList(w)

        if bias:
            self.b = nn.ParameterList(
                [
                    self.scale * torch.randn(2 * self.hidden_size, 1, 1, 2)
                    for _ in range(self.spectral_layers)
                ]
            )

        self.wout = nn.Parameter(
            self.scale * torch.randn(self.hidden_size, self.embed_dim, 2)
        )

        self.drop = nn.Dropout(drop_rate) if drop_rate > 0.0 else nn.Identity()

        self.activation = ComplexReLU(
            mode=complex_activation, bias_shape=(self.hidden_size, 1, 1)
        )

    def forward_mlp(self, xr):  # pragma: no cover
        """forward method for the MLP part of the network"""
        for l in range(self.spectral_layers):
            if hasattr(self, "b"):
                xr = self.mul_add_handle(
                    xr, self.w[l].to(xr.dtype), self.b[l].to(xr.dtype)
                )
            else:
                xr = self.mul_handle(xr, self.w[l].to(xr.dtype))
            xr = torch.view_as_complex(xr)
            xr = self.activation(xr)
            xr = self.drop(xr)
            xr = torch.view_as_real(xr)

        # final MLP
        xr = self.mul_handle(xr, self.wout)

        return xr

    def forward(self, x):  # pragma: no cover
        dtype = x.dtype
        # x = x.to(torch.float32)

        # FWD transform
        with torch.amp.autocast("cuda", enabled=False):
            x = x.to(torch.float32)
            x = x.contiguous()
            x = self.forward_transform(x)
            x = torch.view_as_real(x)

        # MLP
        x = self.forward_mlp(x)

        # BWD transform
        with torch.amp.autocast("cuda", enabled=False):
            x = torch.view_as_complex(x)
            x = x.contiguous()
            x = self.inverse_transform(x)
            x = x.to(dtype)

        return x



import dataclasses
from collections.abc import Mapping, Sequence
from datetime import timedelta
from typing import Literal

import numpy as np
import pandas as pd
import torch
import xarray as xr

from fme.core.typing_ import Slice, TensorDict


@dataclasses.dataclass
class TimeSlice:
    """
    Configuration of a slice of times. Step is an integer-valued index step.

    Note: start_time and stop_time may be provided as partial time strings and the
        stop_time will be included in the slice. See more details in `Xarray docs`_.

    Parameters:
        start_time: Start time of the slice.
        stop_time: Stop time of the slice.
        step: Step of the slice.

    .. _Xarray docs:
       https://docs.xarray.dev/en/latest/user-guide/weather-climate.html#non-standard-calendars-and-dates-outside-the-nanosecond-precision-range
    """  # noqa: E501

    start_time: str | None = None
    stop_time: str | None = None
    step: int | None = None

    def slice(self, time: xr.CFTimeIndex) -> slice:
        return time.slice_indexer(self.start_time, self.stop_time, self.step)


def _convert_interval_to_int(
    interval: pd.Timedelta,
    timestep: timedelta,
):
    """Convert interval to integer number of timesteps."""
    if interval % timestep != timedelta(0):
        raise ValueError(
            f"Requested interval length {interval} is not a "
            f"multiple of the timestep {timestep}."
        )

    return interval // timestep


@dataclasses.dataclass
class RepeatedInterval:
    """
    Configuration for a repeated interval within a block.  This configuration
    is used to generate a boolean mask for a dataset that will return values
    within the interval and repeat that throughout the dataset.

    Parameters:
        interval_length: Length of the interval to return values from
        start: Start position of the interval within the repeat block.
        block_length: Total length of the block to be repeated over the length of
            the dataset, including the interval length.

    Note:
        The interval_length, start, and block_length can be provided as either
        all integers or all strings representing timedeltas of the block.
        If provided as strings, the timestep must be provided when calling
        `get_boolean_mask`.

    Examples:
        To return values from the first 3 items of every 6 items, use:

        >>> fme.ace.RepeatedInterval(interval_length=3, block_length=6, start=0)  # doctest: +IGNORE_OUTPUT

        To return a days worth of values starting after 2 days from every 7-day
        block, use:

        >>> fme.ace.RepeatedInterval(interval_length="1d", block_length="7d", start="2d")  # doctest: +IGNORE_OUTPUT
    """  # noqa: E501

    interval_length: int | str
    start: int | str
    block_length: int | str

    def __post_init__(self):
        types = {type(self.interval_length), type(self.block_length), type(self.start)}
        if len(types) > 1:
            raise ValueError(
                "All attributes of RepeatedInterval must be of the "
                "same type (either all int or all str)."
            )

        self._is_time_delta_str = isinstance(self.interval_length, str)

        if self._is_time_delta_str:
            self.interval_length = pd.Timedelta(self.interval_length)
            self.block_length = pd.Timedelta(self.block_length)
            self.start = pd.Timedelta(self.start)

    def get_boolean_mask(
        self, length: int, timestep: timedelta | None = None
    ) -> np.ndarray:
        """
        Return a boolean mask for the repeated interval.

        Args:
            length: Length of the dataset.
            timestep: Timestep of the dataset.
        """
        if self._is_time_delta_str:
            if timestep is None:
                raise ValueError(
                    "Timestep must be provided when using time deltas "
                    "for RepeatedInterval."
                )

            interval_length = _convert_interval_to_int(self.interval_length, timestep)
            block_length = _convert_interval_to_int(self.block_length, timestep)
            start = _convert_interval_to_int(self.start, timestep)
        else:
            interval_length = self.interval_length
            block_length = self.block_length
            start = self.start

        if start + interval_length > block_length:
            raise ValueError(
                "The interval (with start point) must fit within the repeat block."
            )

        block = np.zeros(block_length, dtype=bool)
        block[start : start + interval_length] = True
        num_blocks = length // block_length + 1
        mask = np.tile(block, num_blocks)[:length]
        return mask


@dataclasses.dataclass
class OverwriteConfig:
    """Configuration to overwrite field values in XarrayDataset.

    Parameters:
        constant: Fill field with constant value.
        multiply_scalar: Multiply field by scalar value.
    """

    constant: Mapping[str, float] = dataclasses.field(default_factory=dict)
    multiply_scalar: Mapping[str, float] = dataclasses.field(default_factory=dict)

    def __post_init__(self):
        key_overlap = set(self.constant.keys()) & set(self.multiply_scalar.keys())
        if key_overlap:
            raise ValueError(
                "OverwriteConfig cannot have the same variable in both constant "
                f"and multiply_scalar: {key_overlap}"
            )

    def apply(self, tensors: TensorDict) -> TensorDict:
        for var, fill_value in self.constant.items():
            data = tensors[var]
            tensors[var] = torch.ones_like(data) * torch.tensor(
                fill_value, dtype=data.dtype, device=data.device
            )
        for var, multiplier in self.multiply_scalar.items():
            data = tensors[var]
            tensors[var] = data * torch.tensor(
                multiplier, dtype=data.dtype, device=data.device
            )
        return tensors

    @property
    def variables(self):
        return set(self.constant.keys()) | set(self.multiply_scalar.keys())


@dataclasses.dataclass
class FillNaNsConfig:
    """
    Configuration to fill NaNs with a constant value or others.

    Parameters:
        method: Type of fill operation. Currently only 'constant' is supported.
        value: Value to fill NaNs with.
    """

    method: Literal["constant"] = "constant"
    value: float = 0.0


@dataclasses.dataclass
class XarrayDataConfig:
    """
    Parameters:
        data_path: Path to the data.
        file_pattern: Glob pattern to match files in the data_path.
        n_repeats: Number of times to repeat the dataset (in time). It is up
            to the user to ensure that the input dataset to repeat results in
            data that is reasonably continuous across repetitions.
        engine: Backend used in xarray.open_dataset call.
        spatial_dimensions: Specifies the spatial dimensions for the grid, default
            is lat/lon. If 'latlon', it is assumed that the last two dimensions are
            latitude and longitude, respectively. If 'healpix', it is assumed that the
            last three dimensions are face, height, and width, respectively.
        subset: Slice defining a subset of the XarrayDataset to load. This can
            either be a `Slice` of integer indices or a `TimeSlice` of timestamps.
            This feature is applied directly to the dataset samples. For example,
            if the file(s) have the time coordinate (t0, t1, t2, t3) and
            requirements.n_timesteps=2, then subset=Slice(stop=2) will
            provide two samples: (t0, t1), (t1, t2).
        infer_timestep: Whether to infer the timestep from the provided data.
            This should be set to True (the default) for ACE training. It may
            be useful to toggle this to False for applications like downscaling,
            which do not depend on the timestep of the data and therefore lack
            the additional requirement that the data be ordered and evenly
            spaced in time. It must be set to True if n_repeats > 1 in order
            to be able to infer the full time coordinate.
        dtype: Data type to cast the data to. If None, no casting is done. It is
            required that 'torch.{dtype}' is a valid dtype.
        overwrite: Optional OverwriteConfig to overwrite loaded field values.
        fill_nans: Optional FillNaNsConfig to fill NaNs with a constant value.
        isel: Optional xarray isel arguments to be passed to the dataset. Will
            raise ValueError if time is included here, since the subset argument
            is used specifically for selecting times. Horizontal dimensions are
            also not currently supported.

    Examples:
        If data is stored in a directory with multiple netCDF files which can be
        concatenated along the time dimension, use:

        >>> fme.ace.XarrayDataConfig(data_path="/some/directory", file_pattern="*.nc") # doctest: +IGNORE_OUTPUT

        If data is stored in a single zarr store at ``/some/directory/dataset.zarr``,
        use:

        >>> fme.ace.XarrayDataConfig(
        ...     data_path="/some/directory",
        ...     file_pattern="dataset.zarr",
        ...     engine="zarr"
        ... ) # doctest: +IGNORE_OUTPUT
    """  # noqa: E501

    data_path: str
    file_pattern: str = "*.nc"
    n_repeats: int = 1
    engine: Literal["netcdf4", "h5netcdf", "zarr"] = "netcdf4"
    spatial_dimensions: Literal["healpix", "latlon"] = "latlon"
    subset: Slice | TimeSlice | RepeatedInterval = dataclasses.field(
        default_factory=Slice
    )
    infer_timestep: bool = True
    dtype: str | None = "float32"
    overwrite: OverwriteConfig = dataclasses.field(default_factory=OverwriteConfig)
    fill_nans: FillNaNsConfig | None = None
    isel: Mapping[str, Slice | int] = dataclasses.field(default_factory=dict)

    def _default_file_pattern_check(self):
        if self.engine == "zarr" and self.file_pattern == "*.nc":
            raise ValueError(
                "The file pattern is set to the default NetCDF file pattern *.nc "
                "but the engine is specified as 'zarr'. Please set "
                "`XarrayDataConfig.file_pattern` to match the zarr filename."
            )

    @property
    def torch_dtype(self) -> torch.dtype | None:
        if self.dtype is None:
            return None
        else:
            try:
                torch_dtype = getattr(torch, self.dtype)
            except AttributeError:
                raise ValueError(f"Invalid dtype '{self.dtype}'")
            if not isinstance(torch_dtype, torch.dtype):
                raise ValueError(f"Invalid dtype '{self.dtype}'")
        return torch_dtype

    def __post_init__(self):
        if self.n_repeats > 1 and not self.infer_timestep:
            raise ValueError(
                "infer_timestep must be True if n_repeats is greater than 1"
            )
        if self.spatial_dimensions not in ["latlon", "healpix"]:
            raise ValueError(
                f"unexpected spatial_dimensions {self.spatial_dimensions},"
                " should be one of 'latlon' or 'healpix'"
            )
        self.torch_dtype  # check it can be retrieved
        self._default_file_pattern_check()
        self.zarr_engine_used = False
        if self.engine == "zarr":
            self.zarr_engine_used = True


@dataclasses.dataclass
class ConcatDatasetConfig:
    """
    Configuration for concatenating multiple datasets.
    Parameters:
        concat: List of XarrayDataConfig objects to concatenate.
        strict: Whether to enforce that the datasets to be concatenated
            have the same dimensions and coordinates.
    """

    concat: Sequence[XarrayDataConfig]
    strict: bool = True

    def __post_init__(self):
        self.zarr_engine_used = any(ds.engine == "zarr" for ds in self.concat)


@dataclasses.dataclass
class MergeDatasetConfig:
    """
    Configuration for merging multiple datasets.
    Parameters:
        merge: List of ConcatDatasetConfig or XarrayDataConfig to merge.
    """

    merge: Sequence[ConcatDatasetConfig | XarrayDataConfig]

    def __post_init__(self):
        self.zarr_engine_used = False
        for ds in self.merge:
            if isinstance(ds, ConcatDatasetConfig):
                if ds.zarr_engine_used:
                    self.zarr_engine_used = ds.zarr_engine_used
                    break
            elif isinstance(ds, XarrayDataConfig):
                if ds.engine == "zarr":
                    self.zarr_engine_used = True
                    break


@dataclasses.dataclass
class MergeNoConcatDatasetConfig:
    """
    Configuration for merging multiple datasets. No concatenation is allowed.
    Parameters:
        merge: List of XarrayDataConfig to merge.
    """

    merge: Sequence[XarrayDataConfig]

    def __post_init__(self):
        self.zarr_engine_used = False
        for ds in self.merge:
            if ds.engine == "zarr":
                self.zarr_engine_used = True
                break



# flake8: noqa
# Copied from https://github.com/NVIDIA/modulus/commit/89a6091bd21edce7be4e0539cbd91507004faf08
# Copyright (c) 2023 - 2024 NVIDIA CORPORATION & AFFILIATES.
# All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import dataclasses
from typing import List, Optional, Sequence

import torch as th
import torch.nn as nn

from .healpix_blocks import ConvBlockConfig, RecurrentBlockConfig


@dataclasses.dataclass
class UNetDecoderConfig:
    """
    Configuration for the UNet Decoder.

    Parameters:
        conv_block: Configuration for the convolutional block.
        up_sampling_block: Configuration for the up-sampling block.
        output_layer: Configuration for the output layer block.
        recurrent_block: Configuration for the recurrent block, by default None.
        n_channels: Number of channels for each layer, by default (34, 68, 136).
        n_layers: Number of layers in each block, by default (1, 2, 2).
        output_channels: Number of output channels, by default 1.
        dilations: List of dilation rates for the layers, by default None.
        enable_nhwc: Flag to enable NHWC data format, by default False.
        enable_healpixpad: Flag to enable HEALPix padding, by default False.
    """

    conv_block: ConvBlockConfig
    up_sampling_block: ConvBlockConfig
    output_layer: ConvBlockConfig
    recurrent_block: Optional[RecurrentBlockConfig] = None
    n_channels: List[int] = dataclasses.field(default_factory=lambda: [34, 68, 136])
    n_layers: List[int] = dataclasses.field(default_factory=lambda: [1, 2, 2])
    output_channels: int = 1
    dilations: Optional[list] = None
    enable_nhwc: bool = False
    enable_healpixpad: bool = False

    def build(self) -> nn.Module:
        """
        Builds the UNet Decoder model.

        Returns:
            UNet Decoder model.
        """
        return UNetDecoder(
            conv_block=self.conv_block,
            up_sampling_block=self.up_sampling_block,
            output_layer=self.output_layer,
            recurrent_block=self.recurrent_block,
            n_channels=self.n_channels,
            n_layers=self.n_layers,
            output_channels=self.output_channels,
            dilations=self.dilations,
            enable_nhwc=self.enable_nhwc,
            enable_healpixpad=self.enable_healpixpad,
        )


class UNetDecoder(nn.Module):
    """Generic UNetDecoder that can be applied to arbitrary meshes."""

    def __init__(
        self,
        conv_block: ConvBlockConfig,
        up_sampling_block: ConvBlockConfig,
        output_layer: ConvBlockConfig,
        recurrent_block: Optional[RecurrentBlockConfig] = None,
        n_channels: Sequence = (64, 32, 16),
        n_layers: Sequence = (1, 2, 2),
        output_channels: int = 1,
        dilations: Optional[list] = None,
        enable_nhwc: bool = False,
        enable_healpixpad: bool = False,
    ):
        """
        Initialize the UNetDecoder.

        Args:
            conv_block: Configuration for the convolutional block.
            up_sampling_block: Configuration for the upsampling block.
            output_layer: Configuration for the output layer.
            recurrent_block: Configuration for the recurrent block. If None, recurrent blocks are not used.
            n_channels: Sequence specifying the number of channels in each decoder layer.
            n_layers: Sequence specifying the number of layers in each block.
            output_channels: Number of output channels.
            dilations: List of dilations to use for the convolutional blocks.
            enable_nhwc: If True, use channel last format.
            enable_healpixpad: If True, use the healpixpad library if installed.
        """
        super().__init__()
        self.channel_dim = 1

        if dilations is None:
            dilations = [1 for _ in range(len(n_channels))]

        self.decoder = []
        for n, curr_channel in enumerate(n_channels):
            up_sample_module = None
            if n != 0:
                up_sampling_block.in_channels = curr_channel
                up_sampling_block.out_channels = curr_channel
                up_sampling_block.enable_nhwc = enable_nhwc
                up_sampling_block.enable_healpixpad = enable_healpixpad
                up_sample_module = up_sampling_block.build()

            next_channel = (
                n_channels[n + 1] if n < len(n_channels) - 1 else n_channels[-1]
            )

            conv_block.in_channels = curr_channel * 2 if n > 0 else curr_channel
            conv_block.latent_channels = curr_channel
            conv_block.out_channels = next_channel
            conv_block.dilation = dilations[n]
            conv_block.n_layers = n_layers[n]
            conv_block.enable_nhwc = enable_nhwc
            conv_block.enable_healpixpad = enable_healpixpad
            conv_module = conv_block.build()

            rec_module = None
            if recurrent_block is not None:
                recurrent_block.in_channels = next_channel
                recurrent_block.enable_healpixpad = enable_healpixpad
                rec_module = recurrent_block.build()

            self.decoder.append(
                nn.ModuleDict(
                    {
                        "upsamp": up_sample_module,
                        "conv": conv_module,
                        "recurrent": rec_module,
                    }
                )
            )

        self.decoder = nn.ModuleList(self.decoder)

        output_layer.in_channels = curr_channel
        output_layer.out_channels = output_channels
        output_layer.dilation = dilations[-1]
        output_layer.enable_nhwc = enable_nhwc
        output_layer.enable_healpixpad = enable_healpixpad

        self.output_layer = output_layer.build()

    def forward(self, inputs):
        """
        Forward pass of the UNetDecoder.

        Args:
            inputs: The inputs to the forward pass.

        Returns:
            The decoded values.
        """
        x = inputs[-1]
        for n, layer in enumerate(self.decoder):
            if layer["upsamp"] is not None:
                up = layer["upsamp"](x)
                x = th.cat([up, inputs[-1 - n]], dim=self.channel_dim)
            x = layer["conv"](x)
            if layer["recurrent"] is not None:
                x = layer["recurrent"](x)
        return self.output_layer(x)

    def reset(self):
        """Resets the state of the decoder layers."""
        for layer in self.decoder:
            if layer["recurrent"] is not None:
                layer["recurrent"].reset()



import abc
from typing import Any, Generic, TypeVar

from torch import nn

from fme.core.generics.inference import PredictFunction
from fme.core.generics.optimization import OptimizationABC
from fme.core.training_history import TrainingJob
from fme.core.typing_ import TensorDict

TO = TypeVar("TO", bound="TrainOutputABC")  # train output


class TrainOutputABC(abc.ABC):
    @abc.abstractmethod
    def get_metrics(self) -> TensorDict:
        pass


PS = TypeVar("PS")  # prognostic state
BD = TypeVar("BD")  # batch data
FD = TypeVar("FD")  # forcing data
SD = TypeVar("SD")  # stepped data


class TrainStepperABC(abc.ABC, Generic[PS, BD, FD, SD, TO]):
    SelfType = TypeVar("SelfType", bound="TrainStepperABC")

    @abc.abstractmethod
    def train_on_batch(
        self,
        data: BD,
        optimization: OptimizationABC,
        compute_derived_variables: bool = False,
    ) -> TO:
        pass

    @property
    @abc.abstractmethod
    def modules(self) -> nn.ModuleList:
        pass

    @abc.abstractmethod
    def get_state(self) -> dict[str, Any]:
        pass

    @abc.abstractmethod
    def load_state(self, state: dict[str, Any]) -> None:
        pass

    @classmethod
    @abc.abstractmethod
    def from_state(cls: type[SelfType], state: dict[str, Any]) -> SelfType:
        pass

    @property
    @abc.abstractmethod
    def n_ic_timesteps(self) -> int:
        pass

    @property
    @abc.abstractmethod
    def predict_paired(self) -> PredictFunction[PS, FD, SD]:
        pass

    def set_eval(self) -> None:
        for module in self.modules:
            module.eval()

    def set_train(self) -> None:
        for module in self.modules:
            module.train()

    @abc.abstractmethod
    def update_training_history(self, training_job: TrainingJob) -> None:
        pass



__version__ = "2025.7.0"


from . import ace
from . import sht_fix as _sht_fix
from .core import Packer, StandardNormalizer, get_device, get_normalizer, using_gpu
from .core.metrics import (
    gradient_magnitude,
    gradient_magnitude_percent_diff,
    rmse_of_time_mean,
    root_mean_squared_error,
    spherical_area_weights,
    time_and_global_mean_bias,
    weighted_mean,
    weighted_mean_bias,
    weighted_mean_gradient_magnitude,
    weighted_nanmean,
    weighted_std,
    weighted_sum,
)

__all__ = [
    "spherical_area_weights",
    "weighted_mean",
    "weighted_mean_bias",
    "root_mean_squared_error",
    "gradient_magnitude",
    "weighted_mean_gradient_magnitude",
    "rmse_of_time_mean",
    "time_and_global_mean_bias",
    "gradient_magnitude_percent_diff",
    "get_device",
    "get_normalizer",
    "Packer",
    "StandardNormalizer",
    "using_gpu",
    "weighted_std",
    "ace",
]



import subprocess
from collections.abc import Sequence
from dataclasses import asdict, dataclass, fields
from typing import overload

from fme.core.wandb import WandB


@dataclass
class TrainingJob:
    """
    Metadata of a stepper training job.

    Parameters:
        git_sha: git sha of the training job
        job_id: wandb run id of the training job
    """

    git_sha: str | None = None
    job_id: str | None = None

    @classmethod
    def from_env(cls) -> "TrainingJob":
        return cls(
            git_sha=git_revision_short_hash(),
            job_id=get_job_id(),
        )


class TrainingHistory(Sequence):
    """Metadata of a stepper's training history.

    Parameters:
        training_jobs: list of training jobs
    """

    def __init__(self, training_jobs: list[TrainingJob] | None = None):
        self._training_jobs: list[TrainingJob] = training_jobs or []

    @overload
    def __getitem__(self, idx: int) -> TrainingJob:
        pass

    @overload
    def __getitem__(self, idx: slice) -> list[TrainingJob]:
        pass

    def __getitem__(self, idx):
        return self._training_jobs[idx]

    def __len__(self) -> int:
        return len(self._training_jobs)

    def __eq__(self, other: object) -> bool:
        if not isinstance(other, TrainingHistory):
            return False
        else:
            return self._training_jobs == other._training_jobs

    def _keys(self) -> set[str]:
        return {field.name for field in fields(TrainingJob)}

    def append(self, job: TrainingJob) -> None:
        """Append a training job to the metadata."""
        self._training_jobs.append(job)

    def extend(self, other_history: "TrainingHistory") -> None:
        """Extend the metadata with another training history."""
        self._training_jobs.extend(other_history)

    def get_history_by_key(self) -> dict[str, list[str]]:
        """Get the history of the training jobs organized by key."""
        return {key: self._get_key_history(key) for key in self._keys()}

    def _get_key_history(self, key: str) -> list[str]:
        return [getattr(job, key) for job in self._training_jobs]

    def get_state(self) -> list[dict[str, str]]:
        return [asdict(job) for job in self._training_jobs]

    @classmethod
    def from_state(cls, state: list[dict[str, str]]) -> "TrainingHistory":
        return cls([TrainingJob(**job) for job in state])


def get_job_id() -> str | None:
    """
    Get the wandb ID of the current job.
    """
    wandb = WandB.get_instance()
    if wandb.enabled and wandb.configured:
        return wandb.get_id()
    else:
        return None


def git_revision_short_hash() -> str | None:
    try:
        return (
            subprocess.check_output(["git", "rev-parse", "--short", "HEAD"])
            .decode("ascii")
            .strip()
        )
    except subprocess.CalledProcessError:
        return None


def get_job_env() -> dict[str, str | None]:
    """
    Get the environment of the current job, including git SHA and wandb job ID.
    """
    return dict(git_sha=git_revision_short_hash(), job_id=get_job_id())



import dataclasses
import datetime

import numpy as np
import pytest
import torch

from fme.ace.stepper import SingleModuleStepperConfig
from fme.core.coordinates import HybridSigmaPressureCoordinate, LatLonCoordinates
from fme.core.dataset_info import DatasetInfo
from fme.core.device import get_device
from fme.core.normalizer import NormalizationConfig

TIMESTEP = datetime.timedelta(hours=6)


@pytest.mark.parametrize("shape", [pytest.param((8, 16))])
def test_sfno_init(shape):
    num_layers = 2
    sfno_config_data = {
        "type": "SphericalFourierNeuralOperatorNet",
        "config": {
            "num_layers": num_layers,
            "embed_dim": 3,
            "scale_factor": 1,
        },
    }
    stepper_config_data = {
        "builder": sfno_config_data,
        "in_names": ["x"],
        "out_names": ["x"],
        "normalization": dataclasses.asdict(
            NormalizationConfig(
                means={"x": float(np.random.randn(1).item())},
                stds={"x": float(np.random.randn(1).item())},
            )
        ),
    }
    horizontal_coordinate = LatLonCoordinates(
        lat=torch.zeros(shape[0]), lon=torch.zeros(shape[1])
    )
    vertical_coordinate = HybridSigmaPressureCoordinate(
        ak=torch.arange(7), bk=torch.arange(7)
    ).to(get_device())
    stepper_config = SingleModuleStepperConfig.from_state(stepper_config_data)
    stepper = stepper_config.get_stepper(
        dataset_info=DatasetInfo(
            horizontal_coordinates=horizontal_coordinate,
            vertical_coordinate=vertical_coordinate,
            timestep=TIMESTEP,
        ),
    )
    assert len(stepper.modules) == 1
    assert len(stepper.modules[0].module.blocks) == num_layers



import datetime

import cftime
import numpy as np
import torch
import xarray as xr

import fme
from fme.ace.aggregator.inference.seasonal import SeasonalAggregator
from fme.core.device import get_device
from fme.core.gridded_ops import LatLonOperations
from fme.core.mask_provider import MaskProvider


def get_zero_time(shape, dims):
    return xr.DataArray(np.zeros(shape, dtype="datetime64[ns]"), dims=dims)


def test_seasonal_aggregator():
    n_lat = 16
    n_lon = 32
    # need to have two actual full years of data for plotting to get exercised
    n_sample = 2
    n_time_step = 8
    n_time = int(365 / 10 * 2 / n_time_step + 1) * n_time_step
    area_weights = torch.ones(n_lat, n_lon).to(fme.get_device())
    agg = SeasonalAggregator(
        LatLonOperations(area_weights),
    )
    target_data = {
        "a": torch.randn(n_sample, n_time, n_lat, n_lon, device=get_device())
    }
    gen_data = {"a": torch.randn(n_sample, n_time, n_lat, n_lon, device=get_device())}

    def time_select(tensor_mapping, start, stop):
        return {
            name: value[:, start:stop, ...] for name, value in tensor_mapping.items()
        }

    time = get_zero_time(shape=[n_sample, n_time], dims=["sample", "time"])
    time_1d = [
        cftime.DatetimeProlepticGregorian(2000, 1, 1) + i * datetime.timedelta(days=10)
        for i in range(n_time)
    ]
    time = xr.DataArray([time_1d for _ in range(n_sample)], dims=["sample", "time"])
    for i in range(0, n_time, n_time_step):
        agg.record_batch(
            time.isel(time=range(i, i + n_time_step)),
            time_select(target_data, i, i + n_time_step),
            time_select(gen_data, i, i + n_time_step),
        )
    logs = agg.get_logs(label="test")
    for name, value in logs.items():
        if isinstance(value, float | np.ndarray):
            assert not np.isnan(value), f"{name} is nan"


def test_seasonal_aggregator_with_nans():
    n_lat = 16
    n_lon = 32
    # need to have two actual full years of data for plotting to get exercised
    n_sample = 2
    n_time_step = 8
    n_time = int(365 / 10 * 2 / n_time_step + 1) * n_time_step
    area_weights = torch.ones(n_lat, n_lon).to(fme.get_device())
    mask = torch.ones((n_lat, n_lon))
    mask[1, 1] = 0
    mask_provider = MaskProvider({"mask_a": mask}).to(get_device())
    agg = SeasonalAggregator(
        LatLonOperations(area_weights, mask_provider),
    )
    target_data = {
        "a": torch.randn(n_sample, n_time, n_lat, n_lon, device=get_device())
    }
    gen_data = {"a": torch.randn(n_sample, n_time, n_lat, n_lon, device=get_device())}
    target_data["a"][:, :, 1, 1] = float("nan")
    gen_data["a"][:, :, 1, 1] = float("nan")

    def time_select(tensor_mapping, start, stop):
        return {
            name: value[:, start:stop, ...] for name, value in tensor_mapping.items()
        }

    time = get_zero_time(shape=[n_sample, n_time], dims=["sample", "time"])
    time_1d = [
        cftime.DatetimeProlepticGregorian(2000, 1, 1) + i * datetime.timedelta(days=10)
        for i in range(n_time)
    ]
    time = xr.DataArray([time_1d for _ in range(n_sample)], dims=["sample", "time"])
    for i in range(0, n_time, n_time_step):
        agg.record_batch(
            time.isel(time=range(i, i + n_time_step)),
            time_select(target_data, i, i + n_time_step),
            time_select(gen_data, i, i + n_time_step),
        )
    logs = agg.get_logs(label="test")
    for name, value in logs.items():
        if isinstance(value, float | np.ndarray):
            assert not np.isnan(value), f"{name} is nan"



import fme

"""
Manually triggered for CI tests on GPU so that tests do not
default to CPU if driver issues prevent use of CUDA.
"""
device = str(fme.get_device())
print(f"Device: {device}")
assert device.startswith("cuda")



import dataclasses
import shutil
import tempfile
from typing import Dict, List, Optional

import click
import dacite
import fsspec
import yaml


def copy(source: str, destination: str):
    """Copy between any two 'filesystems'. Do not use for large files.

    Args:
        source: Path to source file/object.
        destination: Path to destination.
    """
    with fsspec.open(source) as f_source:
        with fsspec.open(destination, "wb") as f_destination:
            shutil.copyfileobj(f_source, f_destination)


@dataclasses.dataclass
class StatsConfig:
    output_directory: str
    beaker_dataset: str
    exclude_runs: List[str] = dataclasses.field(default_factory=list)
    start_date: Optional[str] = None
    end_date: Optional[str] = None


@dataclasses.dataclass
class Config:
    runs: Dict[str, str]
    data_output_directory: str
    stats: StatsConfig


@click.command()
@click.argument("config_yaml", type=str)
def main(config_yaml: str):
    """
    Combine statistics for the data processing pipeline.

    Arguments:
    config_yaml -- Path to the configuration file for the data processing pipeline.
    """
    # imported here so we don't need to install beaker for the tests
    from beaker import Beaker

    with open(config_yaml, "r") as f:
        config_data = yaml.load(f, Loader=yaml.CLoader)
    config = dacite.from_dict(data_class=Config, data=config_data)

    stats_combined_dir = config.stats.output_directory + "/combined/"
    beaker = Beaker.from_env()
    with tempfile.TemporaryDirectory() as tmpdir:
        for filename in (
            "centering.nc",
            "scaling-full-field.nc",
            "scaling-residual.nc",
            "time-mean.nc",
        ):
            copy(stats_combined_dir + filename, tmpdir + "/" + filename)
        runs = [run for run in config.runs if run not in config.stats.exclude_runs]
        run_names = ", ".join(runs)
        if config.stats.start_date is None:
            start = "start of run"
        else:
            start = config.stats.start_date
        if config.stats.end_date is None:
            end = "end of run"
        else:
            end = config.stats.end_date
        beaker.dataset.create(
            config.stats.beaker_dataset,
            tmpdir,
            workspace="ai2/ace",
            description=(
                "Coefficients for normalization for data "
                f"{config.data_output_directory} runs {run_names}. "
                f"Computed from {start} to {end}."
            ),
        )


if __name__ == "__main__":
    main()



import dataclasses
import math
from collections.abc import Callable
from typing import Literal

import torch

from fme.ace.registry.registry import ModuleConfig, ModuleSelector
from fme.core.models.conditional_sfno.sfnonet import Context, ContextConfig
from fme.core.models.conditional_sfno.sfnonet import (
    SphericalFourierNeuralOperatorNet as ConditionalSFNO,
)


def isotropic_noise(
    leading_shape: tuple[int, ...],
    lmax: int,  # length of the ℓ axis expected by isht
    mmax: int,  # length of the m axis expected by isht
    isht: Callable[[torch.Tensor], torch.Tensor],
    device: torch.device,
) -> torch.Tensor:
    # --- draw independent N(0,1) parts --------------------------------------
    coeff_shape = (*leading_shape, lmax, mmax)
    real = torch.randn(coeff_shape, dtype=torch.float32, device=device)
    imag = torch.randn(coeff_shape, dtype=torch.float32, device=device)
    imag[..., :, 0] = 0.0  # m = 0 ⇒ purely real

    # m > 0: make Re and Im each N(0,½)  → |a_{ℓ m}|² has variance 1
    sqrt2 = math.sqrt(2.0)
    real[..., :, 1:] /= sqrt2
    imag[..., :, 1:] /= sqrt2

    # --- global scale that makes Var[T(θ,φ)] = 1 ---------------------------
    scale = math.sqrt(4.0 * math.pi) / lmax  # (Unsöld theorem ⇒ L = lmax)
    alm = (real + 1j * imag) * scale

    return isht(alm)


class NoiseConditionedSFNO(torch.nn.Module):
    def __init__(
        self,
        conditional_model: ConditionalSFNO,
        noise_type: Literal["isotropic", "gaussian"] = "gaussian",
        embed_dim: int = 256,
    ):
        super().__init__()
        self.conditional_model = conditional_model
        self.embed_dim = embed_dim
        self.noise_type = noise_type

    def forward(self, x: torch.Tensor) -> torch.Tensor:
        if self.noise_type == "isotropic":
            lmax = self.conditional_model.itrans_up.lmax
            mmax = self.conditional_model.itrans_up.mmax
            noise = isotropic_noise(
                tuple(x.shape[:-3]) + (self.embed_dim,),
                lmax,
                mmax,
                self.conditional_model.itrans_up,
                device=x.device,
            )
        elif self.noise_type == "gaussian":
            noise = torch.randn(
                [*x.shape[:-3], self.embed_dim, *x.shape[-2:]],
                device=x.device,
                dtype=x.dtype,
            )
        else:
            raise ValueError(f"Invalid noise type: {self.noise_type}")

        embedding_scalar = torch.zeros(
            [*x.shape[:-3], 0], device=x.device, dtype=x.dtype
        )
        return self.conditional_model(
            x, Context(embedding_scalar=embedding_scalar, embedding_2d=noise)
        )


# this is based on the call signature of SphericalFourierNeuralOperatorNet at
# https://github.com/NVIDIA/modulus/blob/b8e27c5c4ebc409e53adaba9832138743ede2785/modulus/models/sfno/sfnonet.py#L292  # noqa: E501
@ModuleSelector.register("NoiseConditionedSFNO")
@dataclasses.dataclass
class NoiseConditionedSFNOBuilder(ModuleConfig):
    """
    Configuration for a noise-conditioned SFNO model.

    Noise is provided as conditioning input to conditional layer normalization.

    Attributes:
        spectral_transform: Type of spherical transform to use.
        filter_type: Type of filter to use.
        operator_type: Type of operator to use.
        residual_filter_factor: Factor by which to downsample the residual.
        embed_dim: Dimension of the embedding.
        noise_embed_dim: Dimension of the noise embedding.
        noise_type: Type of noise to use for conditioning.
        global_layer_norm: Whether to reduce along the spatial domain when applying
            layer normalization.
        num_layers: Number of blocks (SFNO and MLP)in the model.
        use_mlp: Whether to use a MLP in the model.
        mlp_ratio: Ratio of the MLP hidden dimension
            to the embedding dimension.
        activation_function: Activation function to use.
        encoder_layers: Number of encoder layers in the model.
        pos_embed: Whether to use a position embedding.
        big_skip: Whether to use a big skip connection in the model.
        rank: Rank of the model.
        factorization: Factorization to use.
        separable: Whether to use a separable filter.
        complex_network: Whether to use a complex network.
        complex_activation: Activation function to use.
        spectral_layers: Number of spectral layers in the model.
        checkpointing: Whether to use checkpointing.
        filter_residual: Whether to filter residual connections through a
            SHT round-trip. These will always be filtered if residual_filter_factor
            is not 1.
        filter_output: Whether to filter the output of the model through a
            SHT round-trip.
        local_blocks: List of block indices to use discrete-conditional
            convolution (DISCO) blocks, which apply local filters. See
            Ocampo et al. (2022)
            https://arxiv.org/abs/2209.13603 for more details.
    """

    spectral_transform: str = "sht"
    filter_type: str = "non-linear"
    operator_type: str = "diagonal"
    residual_filter_factor: int = 1
    embed_dim: int = 256
    noise_embed_dim: int = 256
    noise_type: Literal["isotropic", "gaussian"] = "gaussian"
    global_layer_norm: bool = False
    num_layers: int = 12
    use_mlp: bool = True
    mlp_ratio: float = 2.0
    activation_function: str = "gelu"
    encoder_layers: int = 1
    pos_embed: bool = True
    big_skip: bool = True
    rank: float = 1.0
    factorization: str | None = None
    separable: bool = False
    complex_network: bool = True
    complex_activation: str = "real"
    spectral_layers: int = 1
    checkpointing: int = 0
    # healpix not supported due to assumptions about number of spatial dims
    data_grid: Literal["legendre-gauss", "equiangular"] = "legendre-gauss"
    filter_residual: bool = False
    filter_output: bool = False
    local_blocks: list[int] | None = None

    def build(
        self,
        n_in_channels: int,
        n_out_channels: int,
        img_shape: tuple[int, int],
    ):
        sfno_net = ConditionalSFNO(
            params=self,
            in_chans=n_in_channels,
            out_chans=n_out_channels,
            img_shape=img_shape,
            context_config=ContextConfig(
                embed_dim_scalar=0,
                embed_dim_2d=self.noise_embed_dim,
            ),
        )
        return NoiseConditionedSFNO(
            sfno_net, noise_type=self.noise_type, embed_dim=self.noise_embed_dim
        )



# flake8: noqa
# Copied from https://github.com/ai2cm/modulus/commit/22df4a9427f5f12ff6ac891083220e7f2f54d229
# Copyright (c) 2023, NVIDIA CORPORATION & AFFILIATES. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import torch


@torch.jit.script
def compl_mul1d_fwd(
    a: torch.Tensor, b: torch.Tensor
) -> torch.Tensor:  # pragma: no cover
    """
    Performs a complex-valued multiplication operation between two 1-dimensional
    tensors.
    """
    ac = torch.view_as_complex(a)
    bc = torch.view_as_complex(b)
    resc = torch.einsum("bix,io->box", ac, bc)
    res = torch.view_as_real(resc)
    return res


@torch.jit.script
def compl_muladd1d_fwd(
    a: torch.Tensor, b: torch.Tensor, c: torch.Tensor
) -> torch.Tensor:  # pragma: no cover
    """
    Performs complex multiplication of two 1-dimensional tensors 'a' and 'b', and then
    adds a third tensor 'c'.
    """
    tmpcc = torch.view_as_complex(compl_mul1d_fwd(a, b))
    cc = torch.view_as_complex(c)
    return torch.view_as_real(tmpcc + cc)


@torch.jit.script
def compl_mul2d_fwd(
    a: torch.Tensor, b: torch.Tensor
) -> torch.Tensor:  # pragma: no cover
    """
    Performs a complex-valued multiplication operation between two 2-dimensional
    tensors.
    """
    ac = torch.view_as_complex(a)
    bc = torch.view_as_complex(b)
    resc = torch.einsum("bixy,io->boxy", ac, bc)
    res = torch.view_as_real(resc)
    return res


@torch.jit.script
def compl_muladd2d_fwd(
    a: torch.Tensor, b: torch.Tensor, c: torch.Tensor
) -> torch.Tensor:  # pragma: no cover
    """
    Performs complex multiplication of two 2-dimensional tensors 'a' and 'b', and then
    adds a third tensor 'c'.
    """
    tmpcc = torch.view_as_complex(compl_mul2d_fwd(a, b))
    cc = torch.view_as_complex(c)
    return torch.view_as_real(tmpcc + cc)


@torch.jit.script  # TODO remove
def _contract_localconv_fwd(
    a: torch.Tensor, b: torch.Tensor
) -> torch.Tensor:  # pragma: no cover
    """
    Performs a complex local convolution operation between two tensors 'a' and 'b'.
    """
    ac = torch.view_as_complex(a)
    bc = torch.view_as_complex(b)
    resc = torch.einsum("bixy,iox->boxy", ac, bc)
    res = torch.view_as_real(resc)
    return res


@torch.jit.script  # TODO remove
def _contract_blockconv_fwd(
    a: torch.Tensor, b: torch.Tensor
) -> torch.Tensor:  # pragma: no cover
    """
    Performs a complex block convolution operation between two tensors 'a' and 'b'.
    """
    ac = torch.view_as_complex(a)
    bc = torch.view_as_complex(b)
    resc = torch.einsum("bim,imn->bin", ac, bc)
    res = torch.view_as_real(resc)
    return res


@torch.jit.script  # TODO remove
def _contractadd_blockconv_fwd(
    a: torch.Tensor, b: torch.Tensor, c: torch.Tensor
) -> torch.Tensor:  # pragma: no cover
    """
    Performs a complex block convolution operation between two tensors 'a' and 'b', and
    then adds a third tensor 'c'.
    """
    tmpcc = torch.view_as_complex(_contract_blockconv_fwd(a, b))
    cc = torch.view_as_complex(c)
    return torch.view_as_real(tmpcc + cc)


# for the experimental layer
@torch.jit.script  # TODO remove
def compl_exp_mul2d_fwd(
    a: torch.Tensor, b: torch.Tensor
) -> torch.Tensor:  # pragma: no cover
    """
    Performs a 2D complex multiplication operation between two tensors 'a' and 'b'.
    """
    ac = torch.view_as_complex(a)
    bc = torch.view_as_complex(b)
    resc = torch.einsum("bixy,xio->boxy", ac, bc)
    res = torch.view_as_real(resc)
    return res


@torch.jit.script
def compl_exp_muladd2d_fwd(  # TODO remove
    a: torch.Tensor, b: torch.Tensor, c: torch.Tensor
) -> torch.Tensor:  # pragma: no cover
    """
    Performs a 2D complex multiplication operation between two tensors 'a' and 'b',
    and then adds a third tensor 'c'.
    """
    tmpcc = torch.view_as_complex(compl_exp_mul2d_fwd(a, b))
    cc = torch.view_as_complex(c)
    return torch.view_as_real(tmpcc + cc)


@torch.jit.script
def real_mul2d_fwd(
    a: torch.Tensor, b: torch.Tensor
) -> torch.Tensor:  # pragma: no cover
    """
    Performs a 2D real multiplication operation between two tensors 'a' and 'b'.
    """
    res = torch.einsum("bixy,io->boxy", a, b)
    return res


@torch.jit.script
def real_muladd2d_fwd(
    a: torch.Tensor, b: torch.Tensor, c: torch.Tensor
) -> torch.Tensor:  # pragma: no cover
    """
    Performs a 2D real multiplication operation between two tensors 'a' and 'b', and
    then adds a third tensor 'c'.
    """
    res = real_mul2d_fwd(a, b) + c
    return res


# new contractions set to replace older ones. We use complex
@torch.jit.script
def _contract_diagonal(
    a: torch.Tensor, b: torch.Tensor
) -> torch.Tensor:  # pragma: no cover
    """
    Performs a complex diagonal operation between two tensors 'a' and 'b'.
    """
    ac = torch.view_as_complex(a)
    bc = torch.view_as_complex(b)
    resc = torch.einsum("bixy,ioxy->boxy", ac, bc)
    res = torch.view_as_real(resc)
    return res


@torch.jit.script
def _contract_dhconv(
    a: torch.Tensor, b: torch.Tensor
) -> torch.Tensor:  # pragma: no cover
    """
    Performs a complex Driscoll-Healy style convolution operation between two tensors
    'a' and 'b'.
    """
    ac = torch.view_as_complex(a)
    bc = torch.view_as_complex(b)
    resc = torch.einsum("bixy,iox->boxy", ac, bc)
    res = torch.view_as_real(resc)
    return res


@torch.jit.script
def _contract_sep_diagonal(
    a: torch.Tensor, b: torch.Tensor
) -> torch.Tensor:  # pragma: no cover
    """
    Performs a complex convolution operation between two tensors 'a' and 'b'
    """
    ac = torch.view_as_complex(a)
    bc = torch.view_as_complex(b)
    resc = torch.einsum("bixy,ixy->boxy", ac, bc)
    res = torch.view_as_real(resc)
    return res


@torch.jit.script
def _contract_sep_dhconv(
    a: torch.Tensor, b: torch.Tensor
) -> torch.Tensor:  # pragma: no cover
    """
    Performs a complex convolution operation between two tensors 'a' and 'b'
    """
    ac = torch.view_as_complex(a)
    bc = torch.view_as_complex(b)
    resc = torch.einsum("bixy,ix->boxy", ac, bc)
    res = torch.view_as_real(resc)
    return res



import pytest
import torch
from torch import nn

import fme


def test_gradient_accumulation_with_detach():
    model = nn.Linear(1, 1).to(fme.get_device())
    optimizer = torch.optim.Adam(model.parameters(), lr=0.001)
    optimizer.zero_grad()
    x = torch.randn(10, 1).to(fme.get_device())
    x_series = []
    for i in range(10):
        new_x = model(x)
        loss = new_x.sum()
        loss.backward()
        x_series.append(new_x)
        x = new_x.detach()
    optimizer.step()


def test_gradient_accumulation_without_detach():
    model = nn.Linear(1, 1).to(fme.get_device())
    optimizer = torch.optim.Adam(model.parameters(), lr=0.001)
    optimizer.zero_grad()
    x = torch.randn(10, 1).to(fme.get_device())
    new_x = model(x)
    loss = new_x.sum()
    loss.backward()
    x = new_x
    new_x = model(x)
    loss = new_x.sum()
    with pytest.raises(RuntimeError) as err:
        loss.backward()
    assert "Trying to backward through the graph a second time" in str(err.value)
    optimizer.step()


def test_gradient_accumulation_with_detach_after_compute():
    model = nn.Linear(1, 1).to(fme.get_device())
    optimizer = torch.optim.Adam(model.parameters(), lr=0.001)
    optimizer.zero_grad()
    x = torch.randn(10, 1).to(fme.get_device())
    x_series = []
    for i in range(10):
        new_x = model(x)
        x_series.append(new_x)
        x = new_x.detach()
    for i in range(len(x_series)):
        loss = x_series[i].sum()
        loss.backward()
    optimizer.step()


def test_gradient_accumulation_with_detach_and_concat():
    model = nn.Linear(1, 1).to(fme.get_device())
    optimizer = torch.optim.Adam(model.parameters(), lr=0.001)
    optimizer.zero_grad()
    x = torch.randn(10, 1).to(fme.get_device())
    x_series = []
    for i in range(10):
        new_x = model(x)
        x_series.append(new_x)
        x = new_x.detach()
    output = torch.stack(x_series, dim=1)
    loss = output[:, 0].sum()
    loss.backward()
    loss = output[:, 1].sum()
    with pytest.raises(RuntimeError) as err:
        loss.backward()
    assert "Trying to backward through the graph a second time" in str(err.value)
